#include <iostream>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <zlib.h>
#include <math.h>
#include <iomanip>
#include <list>
#include <vector>

#include <Inventor/SoInput.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoOffscreenRenderer.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/events/SoMouseButtonEvent.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoResetTransform.h>
#include <Inventor/SoInput.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>
#include <Inventor/nodes/SoResetTransform.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/nodes/SoQuadMesh.h>
#include <Inventor/nodes/SoTextureCoordinate2.h>
#include <Inventor/nodes/SoTextureCoordinateBinding.h>
#include <Inventor/nodes/SoTexture2.h>
#include <Inventor/sensors/SoTimerSensor.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoMaterialBinding.h>
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/SoPickedPoint.h>
#include <Inventor/nodes/SoPickStyle.h>

#include <srfftw.h>

using namespace std;

static char version[80] =
"$Revision: 1.5 $";

// variables
bool animPause = FALSE;
SoSeparator *scene = NULL;
char *textureFilename = "texture.jpg";
char *outFile = "out.iv";
float visc = 0.001;
float dt = 0.01;
float t = 0.0;
int n = 64;
int numParticles = 100;
SbVec3f origin(0.0, 0.0, 0.0);
SbVec3f length(1.0, 1.0, 1.0);
SbVec3f voxel = length/n;
SbVec3f invvoxel = SbVec3f(n/length[0], n/length[1], n/length[2]);
float *u, *u0, *v, *v0, *dens, *dens0, *temp, *temp0;
float tempFloat = 0.0;
float vortEps = 1000.0;
SbVec3f *vortg, *vortp;  // grid and particle based vorticities
SbVec3f *du, *dv;        // velocity derivatives
SoTextureCoordinate2 *texCoord, *texCoord0;
SoSeparator **partSeps = NULL;
float maxSpeed = 0.0;
SoMaterial *planeMat = NULL;
SoSwitch *texSwitch = NULL;
SoSwitch *partSwitch = NULL;
enum {SHOW_FORCE, SHOW_SPEED, SHOW_DENSITY, SHOW_TEXTURE};
int showItem = SHOW_DENSITY;
int showParticles = 1;
int useVorParticles = 0;
const int showItems = 4;
const float ambientTemp = 30.0;
void addSmoke();

// clamp definition
// restricts a number to a range [min, max]
template <class T> inline
T clamp(T x, T min, T max){
  if (x<min)
    return(min);
  else if (x>max)
    return(max);
  else
    return(x);
}

inline float finite(SbVec3f &f){
  return (finite(f[0]) && finite(f[1]) && finite(f[2]));
}

inline float gaussFilt(SbVec3f &x, SbVec3f &xp, float rad){
  float r2 = (xp-x).dot(xp-x);
  return exp(-r2/(2.*rad*rad))/(rad*rad*rad*powf(2.*M_PI,1.5));
}


inline float clampin(const float x, const float a, const float b){

  // sort
  if (a<b){
    if (x<a)
      return(a);
    else if (x>b)
      return(b);
  }
  else {
    if (x<b)
      return(b);
    else if (x>a)
      return(a);
  }

  return(x);

}

static rfftwnd_plan plan_rc, plan_cr;

void init_FFT(int n){
  plan_rc = rfftw2d_create_plan(n, n, FFTW_REAL_TO_COMPLEX, FFTW_IN_PLACE);
  plan_cr = rfftw2d_create_plan(n, n, FFTW_COMPLEX_TO_REAL, FFTW_IN_PLACE);
}

#define FFT(s,u)\
if (s==1) rfftwnd_one_real_to_complex(plan_rc, (fftw_real *)u, (fftw_complex *)u);\
else rfftwnd_one_complex_to_real(plan_cr, (fftw_complex *)u, (fftw_real *)u)

#define floor(x) ((x)>=0.0?((int)(x)):(-((int)(1-(x)))))



// interpolates f at a point t (in [0,1]) between f1 and f2
template <class T> inline
T interp(const T& f0, const T& f1, const T& f2, const T& f3, float t){

  // linear interp
  //return ((1-t)*f1 + t*f2);
		    
  // need dk, dkp1, delk
  T dk, dkp1, delk;
  delk = f2-f1;

  if (delk == 0.0)
    dk = dkp1 = 0.0;
  else {

    // k = 1
    dk = (f2-f0)/2.;
    if (dk * delk <= 0.0)
      dk = 0.0;

    // k = 2
    dkp1 = (f3-f1)/2.;
    if (dkp1 * delk <= 0.0)
      dkp1 = 0.0;
  }

  T a0, a1, a2, a3;

  a3 = dk + dkp1 - 2.*delk;
  a2 = 3.*delk - 2.*dk - dkp1;
  a1 = dk;
  a0 = f1;

  T val = (((a3)*t+a2)*t+a1)*t+a0;
  return (clampin(val,f1,f2));
  
}

inline void pos2idx(float x0, float y0, int &i0, int &j0, int &i1, int &j1,
		    float &s, float &t){
  
  // voxel center space
  // x0 = (x0-origin[0])*invvoxel[0]-0.5;
  // y0 = (y0-origin[1])*invvoxel[1]-0.5;
  x0 = x0*invvoxel[0]-0.5;
  y0 = y0*invvoxel[1]-0.5;

  // voxel index and fraction
  i0 = floor(x0);
  s = x0-i0;
  i0 = (n+(i0%n))%n;
  i1 = (i0+1)%n;

  j0 = floor(y0);
  t = y0-j0;
  j0 = (n+(j0%n))%n;
  j1 = (j0+1)%n;
}

float getVal(float x, float y, const float* u){

  float s, t;
  int i, j, i0, j0, i1, j1;

  pos2idx(x, y, i0, j0, i1, j1, s, t);

  int in = (i0-1+n)%n;
  int i2 = (i1+1)%n;
  int jn = (j0-1+n)%n;
  int j2 = (j1+1)%n;

  // interpolate sideways at jn,j0,j1,j2
  float fjn, fj0, fj1, fj2;
  fjn = interp(u[jn*n+in], u[jn*n+i0], u[jn*n+i1], u[jn*n+i2], s);
  fj0 = interp(u[j0*n+in], u[j0*n+i0], u[j0*n+i1], u[j0*n+i2], s);
  fj1 = interp(u[j1*n+in], u[j1*n+i0], u[j1*n+i1], u[j1*n+i2], s);
  fj2 = interp(u[j2*n+in], u[j2*n+i0], u[j2*n+i1], u[j2*n+i2], s);

  // interpolate vertically
  float ret = interp(fjn, fj0, fj1, fj2, t);

  return(ret);
}

template <class T>
T getVal(float x, float y, const T* u){

  T ret;
  float s, t;
  int i, j, i0, j0, i1, j1;

  pos2idx(x, y, i0, j0, i1, j1, s, t);

  int in = (i0-1+n)%n;
  int i2 = (i1+1)%n;
  int jn = (j0-1+n)%n;
  int j2 = (j1+1)%n;

  // interpolate sideways at jn,j0,j1,j2
  float fjn, fj0, fj1, fj2;
  fjn = interp(u[jn*n+in][0], u[jn*n+i0][0], u[jn*n+i1][0], u[jn*n+i2][0], s);
  fj0 = interp(u[j0*n+in][0], u[j0*n+i0][0], u[j0*n+i1][0], u[j0*n+i2][0], s);
  fj1 = interp(u[j1*n+in][0], u[j1*n+i0][0], u[j1*n+i1][0], u[j1*n+i2][0], s);
  fj2 = interp(u[j2*n+in][0], u[j2*n+i0][0], u[j2*n+i1][0], u[j2*n+i2][0], s);

  // interpolate vertically
  ret[0] = interp(fjn, fj0, fj1, fj2, t);

  // interpolate sideways at jn,j0,j1,j2
  fjn = interp(u[jn*n+in][1], u[jn*n+i0][1], u[jn*n+i1][1], u[jn*n+i2][1], s);
  fj0 = interp(u[j0*n+in][1], u[j0*n+i0][1], u[j0*n+i1][1], u[j0*n+i2][1], s);
  fj1 = interp(u[j1*n+in][1], u[j1*n+i0][1], u[j1*n+i1][1], u[j1*n+i2][1], s);
  fj2 = interp(u[j2*n+in][1], u[j2*n+i0][1], u[j2*n+i1][1], u[j2*n+i2][1], s);
  
  // interpolate vertically
  ret[1] = interp(fjn, fj0, fj1, fj2, t);

  return(ret);
}

void stable_solve(int n, float *u, float *v, float*u0, float *v0,
		  float visc, float dt){
  float x, y, x0, y0, f, r, U[2], V[2], s, t;
  int i, j, i0, j0, i1, j1;

  // add forces times timestep
  for (i=0; i<n*n; i++){
    u[i] += dt*u0[i]; u0[i] = u[i];
    v[i] += dt*v0[i]; v0[i] = v[i];
  }

  // self-advect using semi-Lagrangian
  for (x=0.5/n, i=0; i<n; i++,x+=1.0/n){
    for (y=0.5/n, j=0; j<n; j++,y+=1.0/n){

      x0 = x-dt*u0[i+n*j];
      y0 = y-dt*v0[i+n*j];

      u[j*n+i] = getVal(x0, y0, u0);
      v[j*n+i] = getVal(x0, y0, v0);

    }
  }

  // prepare arrays for FFTW
  for (i=0; i<n; i++)
    for (j=0; j<n; j++){
      u0[i+(n+2)*j] = u[i+n*j];
      v0[i+(n+2)*j] = v[i+n*j];
    }

  // FFT
  FFT(1, u0);
  FFT(1, v0);

  // damp viscosity and conserve mass
  // in fourier space
  for (i=0; i<=n; i+=2){
    x = 0.5*i;
    for (j=0; j<n; j++){
      y = j<=n/2 ? j : j-n;
      r = x*x + y*y;
      if (r==0.0)
	continue;
      f = exp(-r*dt*visc);
      U[0] = u0[i  +(n+2)*j]; V[0] = v0[i  +(n+2)*j];
      U[1] = u0[i+1+(n+2)*j]; V[1] = v0[i+1+(n+2)*j];
      u0[i  +(n+2)*j] = f*( (1-x*x/r)*U[0]     -x*y/r *V[0] );
      u0[i+1+(n+2)*j] = f*( (1-x*x/r)*U[1]     -x*y/r *V[1] );
      v0[i  +(n+2)*j] = f*(   -y*x/r *U[0] + (1-y*y/r)*V[0] );
      v0[i+1+(n+2)*j] = f*(   -y*x/r *U[1] + (1-y*y/r)*V[1] );
    }
  }

  // convert back to spatial domain
  FFT(-1, u0);
  FFT(-1, v0);

  // normalize
  f = 1.0/(n*n);

  for (i=0; i<n; i++)
    for (j=0; j<n; j++){
      u[i+j*n] = f*u0[i+(n+2)*j];
      v[i+n*j] = f*v0[i+(n+2)*j];
    }
  
}


ostream& operator<< (ostream& os, const SbVec3f& x){
  os << setiosflags(ios::showpoint | ios::fixed) << setprecision(8);
  os << "(" << x[0] << ", " << x[1] << ", " << x[2] << ")";
  os << resetiosflags(ios::floatfield) << setprecision(6);
  return(os);
}


void mouseEventHandler(void *data, SoEventCallback *callBack){

  SoEventCallback *eventCB = (SoEventCallback *) data;
  const SoEvent *event = callBack->getEvent();

  static int mouseDown[3];
  static int track = 0;
  static SbVec3f lastPos = SbVec3f(0.,0.,0.);
  static int skip=0;
  
  if (SO_MOUSE_PRESS_EVENT(event, BUTTON1)){
    mouseDown[0] = 1;
    skip = 1;
  }
  else if (SO_MOUSE_RELEASE_EVENT(event, BUTTON1)){
    mouseDown[0] = 0;
  }

  if (SO_MOUSE_PRESS_EVENT(event, BUTTON2)){
    mouseDown[1] = 1;
  }
  else if (SO_MOUSE_RELEASE_EVENT(event, BUTTON2)){
    mouseDown[1] = 0;
  }

  if (mouseDown[0] && mouseDown[1])
    return;

  // add velocity
  if (mouseDown[0]){
    
    const SoPickedPoint *p = eventCB->getPickedPoint();
    if (p){
      
      SbVec3f x = p->getPoint();
      
      if (!skip){
	SbVec3f vec = x-lastPos;
        //cerr << "X :" << x << " Vec: " << vec << "\n";

	int i0, i1, j0, j1;
	float s, t;

	pos2idx(x[0], x[1], i0, j0, i1, j1, s, t);
	if (s>0.5)
	  i0 = i1;
	if (t>0.5)
	  j0 = j1;

	//cerr << "i=" << i0 << " j=" << j0 << " lastPos: " << lastPos << "\n";
	
	// add force
	float forcemul = 500000.0;
	//cerr << "Adding: "
	//     << vec[0]*forcemul << ", "
	//     << vec[1]*forcemul << "           ";
	u0[j0*n+i0] += vec[0]*forcemul;
	v0[j0*n+i0] += vec[1]*forcemul;
      }
      else
	skip = 0;

      lastPos = x;
    }
  }

  // add density
  if (mouseDown[1]){

    const SoPickedPoint *p = eventCB->getPickedPoint();
    if (p){
      
      SbVec3f x = p->getPoint();

      int i0, i1, j0, j1;
      float s, t;

      pos2idx(x[0], x[1], i0, j0, i1, j1, s, t);
      if (s>0.5)
	i0 = i1;
      if (t>0.5)
	j0 = j1;
    
      dens[j0*n+i0] = 1.0;
    }
  }
  
}
				      
void resetTexCoords(){
  for (int j=0; j<n; j++){
    for (int i=0; i<n; i++){
      SbVec2f t = SbVec2f((i+0.5)/n, (j+0.5)/n);
      texCoord->point.set1Value(n*j+i, t);
    }
  }
}


//WW''''''''''''''WWWWW'''''''''''''''''WW''WW''W'''''''''''''WW'WWW'''''''
//WW              WW                    WW  WW  W             WW  WW       
//WW W   WW  W WW WWWW  WW W  WW  WWW  WWWW WW  W  WWW WWW   WWW  WW   WW  
//WWW   WWWW W WW WW    WW W WWWW WW W  WW  WWWWW W WW WW W W WW  WW  WWWW 
//WW W  WW   W WW WW    WW W WW   WW W  WW  WW  W W WW WW W W WW  WW  WW   
//WW  W  WWW  WWW WWWWW  WW   WWW WW W  WW  WW  W  WWW WW W  WWW WWWW  WWW 
//             WW                                                          
//'''''''''''WWW'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
void keyEventHandler(void *data, SoEventCallback *callBack){
  const SoEvent *event = callBack->getEvent();


  // Save scene to file
  if (SO_KEY_PRESS_EVENT(event, W)){

    SoOutput *out = new SoOutput();
    out->openFile(outFile);
    SoWriteAction *wa = new SoWriteAction(out);
    wa->apply(scene);
    out->closeFile();
    cerr << "Dumped scene graph to " << outFile << ".\n";
    delete out;
    delete wa;
  }

  // pause
  if (SO_KEY_PRESS_EVENT(event, SPACE)){
    animPause = !animPause;
    if (animPause)
      cerr << "Paused simulation.\n";
    else
      cerr << "Continuing simulation.\n";
  }
  
  // reset texture coordinates
  if (SO_KEY_PRESS_EVENT(event, R)){
    resetTexCoords();
    cerr << "Reset texture coordinates.\n";
  }

  // show particles
  if (SO_KEY_PRESS_EVENT(event, P)){
    showParticles = !showParticles;
    if (showParticles){
      cerr << "Showing particles.\n";
      partSwitch->whichChild.setValue(0);
    }
    else {
      cerr << "Hiding particles.\n";
      partSwitch->whichChild.setValue(-1);
    }
  }
  
  // show material
  if (SO_KEY_PRESS_EVENT(event, M)){

    showItem = (showItem+1)%showItems;
    
    switch (showItem){

    case SHOW_FORCE:
      cerr << "\nDisplaying force.\n";
      texSwitch->whichChild.setValue(-1);
      break;
      
    case SHOW_SPEED:
      cerr << "\nDisplaying speed.\n";
      texSwitch->whichChild.setValue(-1);
      break;
      
    case SHOW_DENSITY:
      cerr << "\nDisplaying density.\n";
      texSwitch->whichChild.setValue(-1);
      break;
      
    case SHOW_TEXTURE:
      cerr << "\nDisplaying texture.\n";
      resetTexCoords();
      texSwitch->whichChild.setValue(0);
      break;
      
    default:
      break;
    }

  }

}


void animate(void *, SoSensor *){

  if (animPause)
    return;

  // advance time
  t += dt;
  
  // advance fluid
  stable_solve(n, u, v, u0, v0, visc, dt);
  cerr << "\r" << t << "       ";

  // add smoke, temperature
  addSmoke();
  
  // reset forces
  for (int j=0; j<n; j++){
    for (int i=0; i<n; i++){

      const float weight = -9.8;
      const float bouy = 1.0;

      SbVec3f f;
      f[0] = f[2] = 0.0;
      f[1] = weight*dens[j*n+i] + bouy*(temp[j*n+i]-ambientTemp);

      u0[j*n+i] = dt*f[0];
      v0[j*n+i] = dt*f[1];
    }
  }

  // increase vorticity
  // using either vorticity confinement or vortex particles method
  if (!useVorParticles) {  // vorticity confinement

    // compute vorticity
    for (int j=0; j<n; j++){
      for (int i=0; i<n; i++){
      
	int i0 = (i-1+n)%n;
	int j0 = (j-1+n)%n;
	int i1 = (i+1)%n;
	int j1 = (j+1)%n;
      
	vortg[i+j*n][0] = 0.0; //(wj1 - wj0 - v[j*n+i+k1] + v[j*n+i+k0])*0.5*invvoxel[0];
	vortg[i+j*n][1] = 0.0; //(u[j* ...
	vortg[i+j*n][2] = (v[i1+j*n] - v[i0+j*n] - u[i+j1*n] + u[i+j0*n])
	  *0.5*invvoxel[0];
      }
    }

    // add vorticity confinement
    for (int j=0; j<n; j++){
      for (int i=0; i<n; i++){

	int i0 = (i-1+n)%n;
	int j0 = (j-1+n)%n;
	int i1 = (i+1)%n;
	int j1 = (j+1)%n;

	SbVec3f f, eta;
	// n = grad |w|
	eta[0] = (vortg[i1+j*n].length() - vortg[i0+j*n].length()) *
	  0.5*invvoxel[0];
	eta[1] = (vortg[i+j1*n].length() - vortg[i+j0*n].length()) *
	  0.5*invvoxel[1];
	eta[2] = 0.0;//(vortg[i0+j*n].length() + vortg[i1+j*n].length()) * 0.5*invvoxel[0];
	eta.normalize();
     
	f = vortEps*voxel[0]*(eta.cross(vortg[i+j*n]));
      
	u0[j*n+i] += dt*f[0];
	v0[j*n+i] += dt*f[1];
      }
    }

  }
  
  else { // vortex particle method
	
    // compute velocity derivatives
    for (int j=0; j<n; j++){
      for (int i=0; i<n; i++){
      
	int i0 = (i-1+n)%n;
	int j0 = (j-1+n)%n;
	int i1 = (i+1)%n;
	int j1 = (j+1)%n;
      
	du[i+j*n][0] = (u[i1+j*n]-u[i0+j*n])*0.5*invvoxel[0];
	du[i+j*n][1] = (u[i+j1*n]-u[i+j0*n])*0.5*invvoxel[1];
	du[i+j*n][2] = 0.0; // for 2D

	dv[i+j*n][0] = (v[i1+j*n]-v[i0+j*n])*0.5*invvoxel[0];
	dv[i+j*n][1] = (v[i+j1*n]-v[i+j0*n])*0.5*invvoxel[1];
	dv[i+j*n][2] = 0.0; // for 2D
      
      }
    }

    // update particle vorticities
    for (int i=0; i<numParticles; i++){
    
      // particle position
      SoTransform *trans = (SoTransform *) partSeps[i]->getChild(0);
      SbVec3f pos = trans->translation.getValue();
    
      // interpolate derivatives
      // fedkiw uses trilinear
      SbVec3f dup = getVal(pos[0], pos[1], du);
      SbVec3f dvp = getVal(pos[0], pos[1], dv);
    
      dup[2] = dvp[2] = 0.0; // 2D fix only (since getVal is 2D)

      SbVec3f vort(0.0, 0.0, 0.0);
      vort[0] = vortp[i][0]*dup[0] + vortp[i][1]*dup[1] + vortp[i][2]*dup[2];
      vort[1] = vortp[i][0]*dvp[0] + vortp[i][1]*dvp[1] + vortp[i][2]*dvp[2];
      vort[2] = 0.0;//vortp[i][0]*dup[0] + vortp[i][1]*dup[1] + vortp[i][2]*dup[2];

      vortp[i] += dt*vort;
    
    }
  
    // add particles vorticity
    for (int k=0; k<numParticles; k++){

      // normalize vorticity
      vortp[k].normalize();
      vortp[k] *= 1.e-3;

      // particle position
      SoTransform *trans = (SoTransform *) partSeps[k]->getChild(0);
      SbVec3f pos = trans->translation.getValue();
    
      const int radi = 4;
      const float rad = radi*voxel[0];

      int i0 = floor(pos[0]*invvoxel[0]-0.5);    
      int j0 = floor(pos[1]*invvoxel[1]-0.5);

      for (int j=0; j<radi*2; j++){
	for (int i=0; i<radi*2; i++){

	  int i1 = (i0-radi+i+n)%n;
	  int j1 = (j0-radi+j+n)%n;

	  SbVec3f x((i1+0.5)*voxel[0], (j1+0.5)*voxel[1], 0.0f);
	  SbVec3f wp = gaussFilt(x, pos, rad) * vortp[k];
	  SbVec3f Np = pos-x;
	  Np.normalize();
	  SbVec3f Fp = vortEps * Np.cross(vortp[k]);

	  u0[j1*n+i1] += dt*Fp[0];
	  v0[j1*n+i1] += dt*Fp[1];
	}
      }
    }
  }
  

  // advect density
  float x, y, x0, y0;
  for (int j=0; j<n; j++){
    for (int i=0; i<n; i++){

      // voxel center position
      x = origin[0]+voxel[0]*(i+0.5);
      y = origin[1]+voxel[1]*(j+0.5);

      // advected position
      x0 = x-dt*u[i+n*j];
      y0 = y-dt*v[i+n*j];

      dens0[j*n+i] = getVal(x0, y0, dens);
      temp0[j*n+i] = getVal(x0, y0, temp);
    }
  }

  // swap smoke
  float *tmp = dens;
  dens = dens0; dens0 = tmp;

  // swap temperature
  tmp = temp;
  temp = temp0; temp0 = tmp;
  
  
  if (showItem != SHOW_TEXTURE){

    static float maxForce = 0.0;

    // adjust vertex colors
    for (int j=0; j<n; j++){
      for (int i=0; i<n; i++){


	SbColor color;
	
	if (showItem == SHOW_SPEED){
	
	  // speed
	  float speed = v[j*n+i]*v[j*n+i]+u[j*n+i]*u[j*n+i];
	  speed = sqrt(speed);

	  // record speed
	  if (speed>maxSpeed)
	    maxSpeed = speed;

	  float hue = sqrt(speed/maxSpeed);
	  hue = -2.0/3.0*hue+2.0/3.0;

	  color.setHSVValue(hue, 0.8, 0.8);

	}
	else if (showItem == SHOW_FORCE){
	  
	  // force
	  float force = v0[j*n+i]*v0[j*n+i]+u0[j*n+i]*u0[j*n+i];
	  force = sqrt(force);
	  
	  // record force
	  if (force>maxForce)
	    maxForce = force;
	  
	  float hue = sqrt(force/maxForce);
	  hue = -2.0/3.0*hue+2.0/3.0;
	
	  color.setHSVValue(hue, 0.8, 0.8);
	}
	else if (showItem == SHOW_DENSITY){
	  
	  float c = dens[j*n+i];

	  if (c>1.0)
	    c = 1.0;
	  color.setValue(c,c,c);
	  
	}
	
	planeMat->diffuseColor.set1Value(j*n+i, color);
      }
    }

  }

  else {

    // advect texture coordinates
    for (int j=0; j<n; j++){
      for (int i=0; i<n; i++){

	// voxel center position
	x = origin[0]+voxel[0]*(i+0.5);
	y = origin[1]+voxel[1]*(j+0.5);

	// advected position
	x0 = x-dt*u[i+n*j];
	y0 = y-dt*v[i+n*j];

	SbVec2f newCoord = getVal(x0, y0, texCoord->point.getValues(0));
	
	texCoord0->point.set1Value(i+n*j, newCoord);
      }
    }

    // copy texture coordinates
    SbVec2f *coord = texCoord->point.startEditing();
    SbVec2f *coord0 = texCoord0->point.startEditing();
    for (int i=0; i<n*n; i++)
      coord[i] = coord0[i];
    texCoord->point.finishEditing();
    texCoord0->point.finishEditing();

  }

  
  // advect particles
  if (1||showParticles) {
    for (int i=0; i<numParticles; i++){

      SoTransform *trans = (SoTransform *) partSeps[i]->getChild(0);
      SoMaterial *mat = (SoMaterial *) partSeps[i]->getChild(1);
    
      // particle position
      SbVec3f pos = trans->translation.getValue();

      SbVec2f u_s;
      u_s[0] = getVal(pos[0], pos[1], u);
      u_s[1] = getVal(pos[0], pos[1], v);

      // advected position
      pos[0] += dt*u_s[0];
      pos[1] += dt*u_s[1];

      for (int j=0; j<2; j++){
	while (pos[j] > origin[j]+length[j])
	  pos[j] -= length[j];
	while (pos[j] < origin[j])
	  pos[j] += length[j];
      }

      trans->translation.setValue(pos);

      // record speed
      float s = u_s.length();
      if (s>maxSpeed)
	maxSpeed = s;

      // set color
      float hue = sqrt(s/maxSpeed);    // 0->1
      hue = -2.0/3.0*hue+2.0/3.0;      // 0.3333->1.0
      mat->diffuseColor.setHSVValue(hue, 1.0, 0.8);
    }
  }

}


void addSmoke(){
  
  // initialize smoke
  for (int j=0; j<n/30+1; j++)
    for (int i=n/2-7; i<n/2+7; i++)
      dens[j*n+i] = 0.8; //*= sqrt(drand48())*2.0;//1.0;

  // initialize temp
  for (int j=0; j<n/30+1; j++)
    for (int i=n/2-5; i<n/2+5; i++)
      temp[j*n+i] = sqrt(drand48())*200.0;

}


//'''''''''''''WW'''''
//                    
//WWWW W   WWW WW WWW 
//WW WW W W WW WW WW W
//WW WW W W WW WW WW W
//WW WW W  WWW WW WW W
//                    
//''''''''''''''''''''
//
int main(int argc, char *argv[]){

  // read options
  while (1){
    char c = getopt(argc, argv, "d:e:f:hmn:t:v:p:f:");

    if (c == -1)
      break;

    switch (c){

    case 'h':
       cerr
	 << "Usage: " << argv[0]
	 << " [-d dt] [-h] [-o out.iv] [-n res] [-t texture.jpg] [-v visc]\n"
	 << "\n"
	 << "Advects a texture using a 2D fluid solver.\n"
	 << ""
	 << "\n"
	 << " -d num          time step dt (0.01)\n"
	 << " -e num          epsilon to use in vorticity calcs (1000.0)\n"
	 << " -h              this help message\n"
	 << " -o file         name of output scene file (out.iv)\n"
	 << " -m              use vortex particle method\n"
	 << " -n num          resolution of grid, squared (64)\n"
         << " -p num          number of particles to advect (100)\n"
	 << " -t file         use texture contained in file (texture.jpg)\n"
         << " -v num          viscosity of fluid (0.001).\n";
	 
      exit(1);
      break;

    case 'd':
      dt = atof(optarg);
      break;

    case 'e':
      vortEps = atof(optarg);
      break;
      
    case 'f':
      tempFloat = atof(optarg);
      break;

    case 'm':
      useVorParticles = 1;
      break;
      
    case 'n':
      n = atoi(optarg);
      voxel = length/n;
      invvoxel = SbVec3f(n/length[0], n/length[1], n/length[2]);
      break;

    case 'p':
      numParticles = atoi(optarg);
      break;

    case 't':
      textureFilename = optarg;
      break;
       
    case 'v':
      visc = atof(optarg);
      cerr << "Set viscosity to " << visc << endl;
      break;

     default:
       break;
    }
  }
  
  while (optind<argc){
    cerr << "Unknown option: " << argv[optind++] << "\n";
  }
  
  // initialize inventor
  SoXt::init(argv[0]);


  // create scene
  scene = new SoSeparator();

  // create plane sep
  SoSeparator *planeSep = new SoSeparator();
  scene->addChild(planeSep);

  // camera
  //SoOrthographicCamera *camera = new SoOrthographicCamera();
  //camera->position.setValue(origin+length/2.0);
  //scene->addChild(camera);
  
  // complexity
  SoComplexity *comp = new SoComplexity();
  comp->textureQuality.setValue(1.0);
  planeSep->addChild(comp);

  // texture
  SoTexture2 *texture = new SoTexture2();
  texture->filename.setValue(textureFilename);
  texture->model.setValue(SoTexture2::DECAL);

  // put texture in switch
  texSwitch = new SoSwitch();
  texSwitch->addChild(texture);
  texSwitch->whichChild.setValue(showItem==SHOW_TEXTURE?0:-1);
  planeSep->addChild(texSwitch);

  // material
  planeMat = new SoMaterial();
  planeMat->diffuseColor.setNum(n*n);
  for (int i=0; i<n*n; i++)
    planeMat->diffuseColor.set1Value(i, SbColor(0.,0.,0.));
  planeMat->ambientColor.setValue(0.,0.,0.);
  SoMaterialBinding *matBind = new SoMaterialBinding();
  matBind->value.setValue(SoMaterialBinding::PER_VERTEX);
  SoShapeHints *quadHint = new SoShapeHints();
  quadHint->vertexOrdering.setValue(SoShapeHints::COUNTERCLOCKWISE);
  planeSep->addChild(quadHint);
  planeSep->addChild(planeMat);
  planeSep->addChild(matBind);
  
  // add grid
  SoCoordinate3 *coord = new SoCoordinate3();
  coord->point.setNum(n*n);
  texCoord = new SoTextureCoordinate2();
  texCoord0 = new SoTextureCoordinate2();
  texCoord->point.setNum(n*n);
  texCoord0->point.setNum(n*n);
  
  // vertices are at grid voxel centers
  // so are texture coordinates
  // grid is O,O->1,1  voxel size = 1/n
  for (int j=0; j<n; j++){
    for (int i=0; i<n; i++){
      SbVec3f p = origin+SbVec3f(voxel[0]*(i+0.5), voxel[1]*(j+0.5), 0.0);
      SbVec2f t = SbVec2f((i+0.5)/n, (j+0.5)/n);
      coord->point.set1Value(n*j+i, p);
      texCoord->point.set1Value(n*j+i, t);
      texCoord0->point.set1Value(n*j+i, t);
    }
  }
  planeSep->addChild(coord);
  planeSep->addChild(texCoord);

  // pick style
  SoPickStyle *pickStyle = new SoPickStyle();
  pickStyle->style.setValue(SoPickStyle::BOUNDING_BOX);
  planeSep->addChild(pickStyle);

  // add mesh
  SoQuadMesh *quadMesh = new SoQuadMesh();
  quadMesh->verticesPerColumn.setValue(n);
  quadMesh->verticesPerRow.setValue(n);
  planeSep->addChild(quadMesh);

  // complexity
  SoComplexity *comp2 = new SoComplexity();
  comp2->value.setValue(0.2);
  scene->addChild(comp2);
  
  // add particles
  partSeps = new SoSeparator*[numParticles];
  SoSeparator *partSep = new SoSeparator();
  partSwitch = new SoSwitch();
  partSwitch->addChild(partSep);
  partSwitch->whichChild.setValue(showParticles?0:-1);

  for (int i=0; i<numParticles; i++){
    partSeps[i] = new SoSeparator();
    
    SoTransform *trans = new SoTransform();

    // jittered initial positions
    int s = floor(sqrt((float)numParticles));
    int k = i%s;
    int j = (i-k)/s;
    trans->translation.setValue((k+drand48())/s, (j+drand48())/s, 0.0);
    partSeps[i]->addChild(trans);
				  
    SoMaterial *mat = new SoMaterial();
    mat->diffuseColor.setValue(0.0, 1.0, 0.0);
    partSeps[i]->addChild(mat);
    
    SoSphere *sph = new SoSphere();
    sph->radius.setValue(0.01);
    partSeps[i]->addChild(sph);
    
    partSep->addChild(partSeps[i]);
  }
  scene->addChild(partSwitch);

  // Set up keyboard
  SoEventCallback *callBack = new SoEventCallback();
  callBack->addEventCallback(SoKeyboardEvent::getClassTypeId(),
			     keyEventHandler);
  callBack->addEventCallback(SoEvent::getClassTypeId(),
			     mouseEventHandler, callBack);
  scene->addChild(callBack);

  cerr <<
    "\n"
    "Press ESC, then P to pause.\n"
    "\n"
    "Keys:\n"
    "  SPACE  pause simulation\n"
    "    R    reset texture coordinates\n"
    "    W    write scene to file\n"
    "    P    toggle particles\n"
    "    M    toggle displaying force/speed/density/texture\n"
    "\n"
       << endl;


  // initialize velocities, force, density
  u = new float[n*n];
  v = new float[n*n];
  dens = new float[n*n];
  temp = new float[n*n];
  u0 = new float[n*(n+2)];   // slightly larger to accomodate FFTW
  v0 = new float[n*(n+2)];
  du = new SbVec3f[n*n];
  dv = new SbVec3f[n*n];
  dens0 = new float[n*(n+2)];
  temp0 = new float[n*(n+2)];
  vortg = new SbVec3f[n*n];
  vortp = new SbVec3f[numParticles];
  for (int j=0; j<n; j++){
    for (int i=0; i<n; i++){
      u[j*n+i] = v[j*n+i] = 0.0;
      u0[j*n+i] = 0.0;
      v0[j*n+i] = 0.0;
      dens[j*n+i] = dens0[j*n+i] = 0.0;
      temp[j*n+i] = temp0[j*n+i] = ambientTemp;
      vortg[j*n+i] = SbVec3f(0.,0.,0.);
    }
  }

  // initialize vorticities
  for (int i=0; i<numParticles; i++)
    vortp[i] = SbVec3f(0.,0.,drand48()<0.5?1.e-3:-1.e-3);


//  addSmoke();

  // initialize FFTW
  init_FFT(n);

  // This returns a main window to use.
  // If unsuccessful, exit.
  Widget myWindow = SoXt::init(argv[0]); // pass the app name
  if (myWindow == NULL) exit(1);

  // timer
  SoTimerSensor *sensor = new SoTimerSensor(animate, texCoord);
  sensor->setInterval(SbTime(0.033)); // 30 frames per second
  sensor->schedule();

  // Set up viewer
  SoXtExaminerViewer *myViewer = new SoXtExaminerViewer(myWindow);
  myViewer->setSceneGraph(scene);
  //myViewer->setCamera(camera);
  char title[20] = "";
  sprintf(title, "Fluid (e=%g)", vortEps);
  myViewer->setTitle(title);
  myViewer->show();
    
  SoXt::show(myWindow); // Display main window
  SoXt::mainLoop();	// Main Inventor event loop
  
  return(0);
}
