//WWWW''''''''WW''''''''WWW''''''''WW''''''''''WWW''''''''''''''''''
//WW  W       WW       WW  W       WW           WW                  
//WW  W  WWW WWWW  WWW  WW    WW  WWWW     WWW  WW   WWW  WWWW  WWWW
//WW  W W WW  WW  W WW   WW  WWWW  WW     WW W  WW  W WW WWW   WWW  
//WW  W W WW  WW  W WW W  WW WW    WW     WW    WW  W WW   WWW   WWW
//WWWW   WWW  WW   WWW  WWW   WWW  WW      WWW WWWW  WWW WWWW  WWWW 
//                                                                  
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

#include "DataSet.h"
#include "debug.h"
#include "Ray.h"
#include "Spectrum.h"
#include "Noise.h"

    /*     ____
       /|4  /|6    z
      / |  / |     |
     /__|0=min2    |__y
    5|  / |7=max  /
     | /  | /    x
    1|/___|3

          012
          ---
    0 = p 000
    1 = p 100
    2 = p 010
    3 = p 110
    4 = p 001
    5 = p 101
    6 = p 011
    7 = p 111

  */

const int DataSet::colorSamples = 1000;
const float DataSet::minTemp = 900.0f;
const float DataSet::maxTemp = 10000.0f;
const float DataSet::tempDeltaInv = 1.0f/((maxTemp-minTemp)/(colorSamples-1));
  
// constructor
DataSet::DataSet(){
  data = NULL;
  colors = NULL;
  Fx = Fy = Fz = NULL;
  min = SbVec3f(0.0, 0.0, 0.0);
  max = SbVec3f(100.0, 100.0, 100.0);
  dim[0] = 100;
  dim[1] = 100;
  dim[2] = 100;
  voxel[0] = (max[0]-min[0])/dim[0];
  voxel[1] = (max[1]-min[1])/dim[1];
  voxel[2] = (max[2]-min[2])/dim[2];
  voxelVol = voxel[0]*voxel[1]*voxel[2];
  invVoxel[0] = dim[0]/(max[0]-min[0]);
  invVoxel[1] = dim[1]/(max[1]-min[1]);
  invVoxel[2] = dim[2]/(max[2]-min[2]);
  diffColor = SbColor(0.7, 0.7, 0.7);
  specColor = SbColor(0.0, 0.0, 0.0);
  shininess = 0;
  transparency = 0.0;
  minValue = 1e30;
  maxValue = -1e30;
  totalDensity = 0.0;
  totalBrightness = 0.0;
  transform.makeIdentity();
}

float DataSet::val(const SbVec3f &x_) const {

  // constant
  //return (0.5);
  
  SbVec3f x;
  transform.multVecMatrix(x_, x);
  
  // convert position into index+offset
  int i, j, k;
  float a, b, c;
  xToijkabc(x, i, j, k, a, b, c);

  int i1 = i+1;
  int j1 = j+1;
  int k1 = k+1;
  
  // interpolate
  float f =
    (1.0f-a)*(1.0f-b)*(1.0f-c) * val( i,  j,  k) +
    (1.0f-a)*(1.0f-b)*(c)      * val( i,  j, k1) +
    (1.0f-a)*(b)*(1.0f-c)      * val( i, j1,  k) +
    (1.0f-a)*(b)*(c)           * val( i, j1, k1) +
    
    (a)*(1.0f-b)*(1.0f-c)      * val(i1,  j,  k) +
    (a)*(1.0f-b)*(c)           * val(i1,  j, k1) +
    (a)*(b)*(1.0f-c)           * val(i1, j1,  k) +
    (a)*(b)*(c)                * val(i1, j1, k1);

  //static Noise h;
  //f *= h.val(x/5.0f, 0.5, .85);

  return(f);
}

// this interpolates but doesn't transformn
float DataSet::valInterpNoX(const SbVec3f &x) const {
  
  // convert position into index+offset
  int i, j, k;
  float a, b, c;
  xToijkabc(x, i, j, k, a, b, c);

  int i1 = i+1;
  int j1 = j+1;
  int k1 = k+1;
  
  // interpolate
  float f =
    (1.0f-a)*(1.0f-b)*(1.0f-c) * val( i,  j,  k) +
    (1.0f-a)*(1.0f-b)*(c)      * val( i,  j, k1) +
    (1.0f-a)*(b)*(1.0f-c)      * val( i, j1,  k) +
    (1.0f-a)*(b)*(c)           * val( i, j1, k1) +
    
    (a)*(1.0f-b)*(1.0f-c)      * val(i1,  j,  k) +
    (a)*(1.0f-b)*(c)           * val(i1,  j, k1) +
    (a)*(b)*(1.0f-c)           * val(i1, j1,  k) +
    (a)*(b)*(c)                * val(i1, j1, k1);

  //static Noise h;
  //f *= h.val(x/5.0f, 0.5, .85);

  return(f);
}

float DataSet::valPoint(const SbVec3f &x_) const {

  SbVec3f x;
  transform.multVecMatrix(x_, x);
  
  // convert position into index
  int i, j, k;
  xToijk(x, i, j, k);

  return data[Fx[i]+Fy[j]+Fz[k]];
}

SbVec4f DataSet::calcBaryCent(SbVec3f x){

  cerr << "DataSet::calcBaryCent() broken due to change to voxel centers\n";

  // convert position into index+offset
  int i, j, k;
  float a, b, c;
  xToijkabc(x, i, j, k, a, b, c);

  return(SbVec4f(a, b, c, 0.0 ));
}

SbVec3f DataSet::getGradient(SbVec3f x) const {

  cerr << "DataSet::getGradient() broken due to change to voxel centers\n";

  // convert position into index+offset
  int t, u, v;
  float q, r, s;
  xToijkabc(x, t, u, v, q, r, s);
  
  SbVec3f n = SbVec3f(0.0, 0.0, 0.0);
  SbVec3f a[2];
  a[1] = SbVec3f(q, r, s);
  a[0] = SbVec3f(1.0-q, 1.0-r, 1.0-s);

  float p_ijk;
    
  for (int i=0; i<2; i++){
    for (int j=0; j<2; j++){
      for (int k=0; k<2; k++){
	  
	p_ijk = val(t+i, u+j, v+k);
	
	n[0] += (i==0?-1.0:1.0) * a[j][1] * a[k][2] * p_ijk;
	n[1] += (j==0?-1.0:1.0) * a[i][0] * a[k][2] * p_ijk;
	n[2] += (k==0?-1.0:1.0) * a[i][0] * a[j][1] * p_ijk;

      }
    }
  }

  n[0] *= invVoxel[0];
  n[1] *= invVoxel[1];
  n[2] *= invVoxel[2];

  return(n);
}

SbVec3f DataSet::getIntNormal(SbVec3f x) const {

  SbVec3f n = getGradient(x);
  n.normalize();
  
  return(n);
}


// read file into temporary memory block,
// pass to other useData() for further processing.
void DataSet::useData(const string &filename, int _dim[],
		      int header, int fourBytes,
		      SbVec3f _min, SbVec3f _max){

  // store min and max
  min = _min;
  max = _max;

  // store dims
  dim[0] = _dim[0];
  dim[1] = _dim[1];
  dim[2] = _dim[2];

  // compute deltas (voxel size)
  // x1-x0
  voxel[0] = (max[0]-min[0])/(dim[0]);
  voxel[1] = (max[1]-min[1])/(dim[1]);
  voxel[2] = (max[2]-min[2])/(dim[2]);

  defaultStepSize = MIN2(voxel[0], MIN2(voxel[1], voxel[2]));
  //defaultStepSize *= .1;
  
  // 1/(x1-x0)
  invVoxel[0] = (dim[0])/(max[0]-min[0]);
  invVoxel[1] = (dim[1])/(max[1]-min[1]);
  invVoxel[2] = (dim[2])/(max[2]-min[2]);

  voxelVol = voxel[0]*voxel[1]*voxel[2];

  // calculate dimensions of data
  int numPts = dim[0]*dim[1]*dim[2];

  // read data into temporary memory block
  float *fdata = new float[numPts];

  if (filename.empty()){

    // create test data
    SbVec3f center = 0.5*(min+max);
    SbVec3f size = 0.5*(max-min);
    
    for (int z=0; z<dim[2]; z++)
      for (int y=0; y<dim[1]; y++)
        for (int x=0; x<dim[0]; x++){
          int p = dim[0]*dim[1]*z + dim[0]*y + x;
          SbVec3f px = ijkabcToPos(x, y, z, 0,0,0);
          SbVec3f ofs = px-center;
          SbVec3f ofsa = ofs;
          ofsa[0] = abs(ofs[0])/float(size[0]);
          ofsa[1] = abs(ofs[1])/float(size[1]);
          ofsa[2] = abs(ofs[2])/float(size[2]);
          float dist = ofsa.length();
          fdata[p] = dist>1 ? 0 : 1-dist;
          fdata[p] = clamp(fdata[p], 0.f, 1.f);
//          fdata[p] = (dim[0]-1-x)/float(dim[0]-1);
          
//           if (ofs[0]>0 && ofs[1]>0 && ofs[2]<0 ||
//               ofs[0]<0 && ofs[1]>0 && ofs[2]>0 ||
//               ofs[0]>0 && ofs[1]<0 && ofs[2]>0 ||
//               ofs[0]<0 && ofs[1]<0 && ofs[2]<0)
//             fdata[p] = 0;
        }

  } else {
    
    // read file into temporary memory
    FILE *fp = fopen(filename.c_str(), "r");
    if (!fp) {
      cerr << "Error: couldn't read " << filename << endl;
      delete [] fdata;
      exit(1);
    }

    if (fseek(fp, header, SEEK_SET)) {
      cerr << "Error: couldn't seek file " << filename << endl;
      delete [] fdata;
      exit(1);
    }

    if (fourBytes == 0){
      unsigned char *ucdata = new unsigned char[numPts];
      if (fread(ucdata, sizeof(unsigned char), numPts, fp) != numPts) {
        delete [] fdata;
        delete [] ucdata;
        cerr << "Error: couldn't read all " << numPts << " bytes of data"
             << endl;
        exit(1);
      }

      for (int i = 0; i < numPts; i++)
        fdata[i] = (float)ucdata[i];
      delete [] ucdata;
    
    }
    // 4 bytes
    else {
      if (fread(fdata, sizeof(float), numPts, fp) != numPts) {
        delete [] fdata;
        cerr << "Error: couldn't read all " << numPts*4 << " bytes of data"
             << endl;
        exit(1);
      }

      // assume data stored as little-endian as a impromptu standard
      // reverse bytes if on big endian machine

      /* Are we little or big endian?  From Harbison&Steele.  */
      union
      {
        long l;
        char c[sizeof (long)];
      } u;
      u.l = 1;
      int bigEndian = (u.c[sizeof (long) - 1] == 1);

      if (bigEndian){
      
        cerr << "Big-endian machine detected: reversing bytes.\n";
        for (int j=0; j<numPts; j++)
          for (int i=0; i<sizeof(float)/2; i++){
            char *c = (char *)&fdata[j];
            char tmp = c[i];
            c[i] = c[sizeof(float)-1-i];
            c[sizeof(float)-1-i] = tmp;
          }
      }
    
    }
  }

  // copy data
  const int mn = m*n;
  int Nx=dim[0], Ny=dim[1], Nz=dim[2];
  int Xmn=Nx/mn, Ymn=Ny/mn, Zmn=Nz/mn;
  if (Xmn*mn<Nx)
    Xmn++;
  if (Ymn*mn<Ny)
    Ymn++;
  if (Zmn*mn<Nz)
    Zmn++;
  sdim[0] = Nx = mn*Xmn;
  sdim[1] = Ny = mn*Ymn;
  sdim[2] = Nz = mn*Zmn;

  cerr << "Nx=" << Nx
       << " Ny=" << Ny
       << " Nz=" << Nz << " ";
  
  // precompute index tables
  Fx = new int[dim[0]];
  Fy = new int[dim[1]];
  Fz = new int[dim[2]];

  for (int x=0; x<dim[0]; x++)
    Fx[x] =
      ((x/n)/m)* n*n*n* m*m*m* ((Nz/n)/m)* ((Ny/n)/m) +
      ((x/n)%m)* n*n*n* m*m +
      (x%n)* n*n;
      
  for (int y=0; y<dim[1]; y++)
    Fy[y] =
      ((y/n)/m)* n*n*n* m*m*m* ((Nz/n)/m) +
      ((y/n)%m)* n*n*n* m +
      (y%n)* n;

  for (int z=0; z<dim[2]; z++)
    Fz[z] =
      ((z/n)/m)* n*n*n* m*m*m +
      ((z/n)%m)* n*n*n +
      (z%n);
      
  // allocate space
  cerr << "Allocating " << Nx*Ny*Nz << " floats \n";
  data = new float[Nx*Ny*Nz];
  int maxq = 0;

  // copy data
  float avg=0.0, nonZeroAvg=0.0;
  int nonZeroVals=0;
  for (int z=0; z<dim[2]; z++)
    for (int y=0; y<dim[1]; y++)
      for (int x=0; x<dim[0]; x++){
	int q = Fx[x] + Fy[y] + Fz[z];

	// ((x/n)/m)* n*n*n* m*m*m* ((d2/n)/m)* ((d1/n)/m) +
	// ((y/n)/m)* n*n*n* m*m*m* ((d2/n)/m) +
	// ((z/n)/m)* n*n*n* m*m*m +
	// ((x/n)%m)* n*n*n* m*m +
	// ((y/n)%m)* n*n*n* m +
	// ((z/n)%m)* n*n*n +
	// (x%(n*n))* n*n +
	// (y%n)* n +
	// (z%n);

	if (q>maxq)
	  maxq = q;
	int p = dim[0]*dim[1]*z + dim[0]*y + x;
	// cerr << "Storing " << p << " into " << q << "\n";
	data[q] = fdata[p];

	// record max/min value
	if (data[q]>maxValue)
	  maxValue = data[q];
	if (data[q]<minValue)
	  minValue = data[q];
	totalDensity += data[q];
	avg += data[q];
	if (data[q]){
	  nonZeroAvg += data[q];
	  nonZeroVals++;
	}
      }

  avg /= dim[0]*dim[1]*dim[2];
  nonZeroAvg /= nonZeroVals;
  
  cerr << "Read 3d scalar data: " << filename << "\n"
       << "dataRange=[" << minValue << ", " << maxValue << "] "
    //<< "Max q=" << maxq
       << "\n"
       << "Avg=" << avg << " Non-zero avg=" << nonZeroAvg << "\n";
  
  // delete temporary storage
  delete [] fdata;
}

// scale all values 
void DataSet::scaleData(float mul){

  cerr << "Scaled data by " << mul << ": "
       << "range=[" << minValue << ", " << maxValue << "] ";

  minValue = 1e30;
  maxValue = -1e30;
  totalDensity = 0.0;
  
  for (int k=0; k<dim[2]; k++){
    for (int j=0; j<dim[1]; j++){
      for (int i=0; i<dim[0]; i++){

	int q = Fx[i] + Fy[j] +Fz[k];
	data[q] *= mul;
	
	if (data[q]>maxValue)
	  maxValue = data[q];
	if (data[q]<minValue)
	  minValue = data[q];
	totalDensity += data[q];
      }
    }
  }

  cerr << "-> "
       << "range=[" << minValue << ", " << maxValue << "]\n";
}

// scale densitites to range [0, 1.0]
void DataSet::scaleMaxToOne(){

  if (maxValue != 0.0)
    scaleData(1.0/maxValue);
}

void DataSet::computeWeights(float scale){

  // x[i,j,k]
  //   r c p
  //       
  //
  //   j
  //   |_ i
  //  /
  // k

  weightsX = new float**[dim[2]];   // x weights P(x|y,z)
  weightsY = new float* [dim[2]];   // y weights P(y|z)
  weightsZ = new float  [dim[2]];   // z weights P(z)
  
  for (int k=0; k<dim[2]; k++){

    weightsX[k] = new float*[dim[1]];
    weightsY[k] = new float [dim[1]];
    
    for (int j=0; j<dim[1]; j++)
      weightsX[k][j] = new float[dim[0]];
  }

  // loop over x,y,z
  float sumX, sumY, sumZ;

  sumZ = 0.0f;
  for (int k=0; k<dim[2]; k++){

    sumY = 0.0f;
    for (int j=0; j<dim[1]; j++){
      
      sumX = 0.0f;
      for (int i=0; i<dim[0]; i++){

	// get blackbody color
	float temp = val(i,j,k)*scale;
	SbColor emisCol = Spectrum(temp).getRGB()*(1.0f/(4.0f*M_PI));
	float bri = brightness(emisCol);
	
	sumX += bri;
	weightsX[k][j][i] = bri;
      }

      weightsY[k][j] = sumX;
      sumY += sumX;

      // normalize x weights
      if (sumX>0.0f)
	for (int i=0; i<dim[0]; i++)
	  weightsX[k][j][i] *= (1.0f/sumX);

    }

    weightsZ[k] = sumY;
    sumZ += sumY;
    
    // normalize y weights
    if (sumY>0.0f)
      for (int j=0; j<dim[1]; j++)
	weightsY[k][j] *= (1.0f/sumY);

  }

  // normalize y weights
  if (sumZ>0.0f)
    for (int k=0; k<dim[2]; k++)
      weightsZ[k] *= (1.0f/sumZ);

  totalBrightness = sumZ;
  cerr << "Total volume brightness = " << totalBrightness << "\n";
}

SbVec3f DataSet::impSample(float x, float r1, float r2, float r3){

  // sanity check
  if (totalBrightness<=0.0f || x>=1.0f){
    cerr << "Error in DataSet::chooseVoxel(): totalBrightness="
	 << totalBrightness << " x=" << x << "\n";
    return min;
  }

  //  2  .3  .3-.9
  //  1  .3  .3-.6
  //  0  .3  .0-.3

  int a, b, c;
  
  // sample P(z)
  float sum = 0.0f;
  c = 0;
  while (sum<x)
    sum += weightsZ[c++];
  c--;

  // sample P(y|z)
  x = (x-(sum-weightsZ[c]))/weightsZ[c];
  sum = 0.0f;
  b = 0;
  while (sum<x)
    sum += weightsY[c][b++];
  b--;

  // sample P(x|y,z)
  x = (x-(sum-weightsY[c][b]))/weightsY[c][b];
  sum = 0.0f;
  a = 0;
  while (sum<x)
    sum += weightsX[c][b][a++];
  a--;

  return ijkabcToPos(a, b, c, r1, r2, r3);
}

  /*     ____
       /|4  /|6    z
      / |  / |     |
     /__|0=min2    |__y
    5|  / |7=max  /
     | /  | /    x
    1|/___|3

          012
          ---
    0 = p 000
    1 = p 100
    2 = p 010
    3 = p 110
    4 = p 001
    5 = p 101
    6 = p 011
    7 = p 111

  */


void DataSet::drawGrid(){
  const float cylRad = (max[X]-min[X])/dim[X]/20.0;
  static SoSeparator *gridSep = NULL;

  // initialize separator
  if (gridSep==NULL){
    gridSep = new SoSeparator();
    scnRoot->addChild(gridSep);
  }
  else {
    gridSep->removeAllChildren();
  }
  
  // draw X axis parallel columns
  SoCylinder *cyl = new SoCylinder();
  cyl->height.setValue(max[X]-min[X]);
  cyl->radius.setValue(cylRad);
  SoTransform *cylTrans = new SoTransform();
  cylTrans->rotation.setValue(0.0, 0.0, 1.0, -1.5707963267949);
  SoSeparator *xCyl = new SoSeparator();
  xCyl->addChild(cylTrans);
  cylTrans = new SoTransform();
  cylTrans->translation.setValue(0.0, (max[X]-min[X])/2.0, 0.0);
  xCyl->addChild(cylTrans);
  xCyl->addChild(cyl);
  
  for (int j=0; j<dim[Y]; j++){
    for (int k=0; k<dim[Z]; k++){
      SoSeparator *bar = new SoSeparator();
      SoTransform *tran = new SoTransform();
      tran->translation.setValue(min[X],
				 j*voxel[Y]+min[Y],
				 k*voxel[Z]+min[Z]);
      bar->addChild(tran);
      bar->addChild(xCyl);
    }
  }

}





//  3D line drawing adapted from Graphics Gem I
//
//  digline: draw digital line from (x1,y1) to (x2,y2),
//  calling a user-supplied procedure at each pixel.
//  Does no clipping.  Uses Bresenham's algorithm.
//  Paul Heckbert3 Sep 85
#define ZSGN(a) ( ((a)<0) ? -1 : (a)>0 ? 1 : 0 )
#define ABS(a) ( ((a)<0) ? -(a) : (a) )

float DataSet::bresenSum(const SbVec3f &p1_, const SbVec3f &p2_,
			 float *t, const float stopAmt){


//   float ret = normalSum(p1_, p2_, t, stopAmt);
//   if (ret<0)
//     SAY_HELLO;
//   return ret;
  
  if (stopAmt<=0){
    *t = 0;
    return 0;
  }
  
  // transform vectors into dataset space
  SbVec3f p1, p2;
  transform.multVecMatrix(p1_, p1);
  transform.multVecMatrix(p2_, p2);
  
  // intersect line with box
  SbVec3f r = p2-p1;
  float t1=0.0, t2=r.normalize();
  SbVec3f tmin, tmax;
  float eps = 1e-25;  // to prevent divide by zero
  tmin[0] = (min[0]-p1[0])/(r[0]+eps); // x=x0+rt  ->  t=(x-x0)/r
  tmax[0] = (max[0]-p1[0])/(r[0]+eps); 
  tmin[1] = (min[1]-p1[1])/(r[1]+eps); 
  tmax[1] = (max[1]-p1[1])/(r[1]+eps); 
  tmin[2] = (min[2]-p1[2])/(r[2]+eps); 
  tmax[2] = (max[2]-p1[2])/(r[2]+eps);

  // swap min/max
  float tmp;
  if (r[0]<0.0){ tmp = tmin[0]; tmin[0] = tmax[0]; tmax[0] = tmp; }
  if (r[1]<0.0){ tmp = tmin[1]; tmin[1] = tmax[1]; tmax[1] = tmp; }
  if (r[2]<0.0){ tmp = tmin[2]; tmin[2] = tmax[2]; tmax[2] = tmp; }

  float mint = MAX3(tmin[0], tmin[1], tmin[2]);
  float maxt = MIN3(tmax[0], tmax[1], tmax[2]);

  // clip p1 and p2 to box
  if (mint>t1) t1 = mint;
  if (maxt<t2) t2 = maxt;

  float sum = 0.0;
  if (t1>=t2)
    return sum;
  
  float dt = t2-t1;

  // constant
//   float cres = 0.5f*dt;
//   if (t && cres>=stopAmt) *t = t1+dt*stopAmt/cres;
//   return (cres);

  int x1, y1, z1, x2, y2, z2;
  float q[2], rr[2], s[2];
  xToijkabc(p1+t1*r, x1, y1, z1, q[0], rr[0], s[0]);
  xToijkabc(p1+t2*r, x2, y2, z2, q[1], rr[1], s[1]);
  
  int xd, yd, zd, x, y, z, ax, ay, az, sx, sy, sz, dx, dy, dz;

  dx = x2-x1;  ax = ABS(dx)<<1;  sx = ZSGN(dx);
  dy = y2-y1;  ay = ABS(dy)<<1;  sy = ZSGN(dy);
  dz = z2-z1;  az = ABS(dz)<<1;  sz = ZSGN(dz);

  x = x1;
  y = y1;
  z = z1;
  float tau = 0.0, ptau = 0.0, v = val(x,y,z), lastv = 0.0;

  // tau = Int_t( sigma_t dt )
  
  if (ax>=ay && ax>=az) {      // x dominant

//     float dxt[3]; // first cell dx, last cell dx, total dx (all positive)
//     if (sx>0){
//       dxt[0] = 1.f-q[0];
//       dxt[1] = q[1];
//       dxt[2] = dxt[0]+dxt[1]+dx-1;
//     }
//     else {
//       dxt[0] = q[0];
//       dxt[1] = 1.f-q[1];
//       dxt[2] = dxt[0]+dxt[1]-dx-1;
//     }
    
    yd = ay-(ax>>1);
    zd = az-(ax>>1);
    for (;;) {

//       v = val(x,y,z);      
//       if (dx==0){
// 	float dtau = dt*v;
// 	tau += dtau;
// 	sum += dtau*v;
//       }
//       else if (x==x0){
// 	float dtau = 
//       }
      
      float prevTau = tau;
      v = val(x,y,z);
      sum += (x==x1) ? 0.0f : (v+lastv)*0.5f;
      lastv = v;
//      tau = (dx==0) ? dt*v : sum/(dx*sx)*dt;
      tau = (dx==0) ? dt*v : sum/float((x-x1+sx)*sx)*dt;
      if (x==x2 || tau>=stopAmt){
	if (t){
	  float pcent = (tau>=stopAmt) ? (stopAmt-prevTau)/(tau-prevTau) : 1.0f;
	  *t = dx==0 ? t1 + pcent*dt : 
            t1 + ( (tau>0.0f) ? ( dt * (x-x1-sx+sx*pcent)/(float)dx) : 0.0f );
	}
	return tau;
      }
      if (yd>=0) {
	y += sy;
	yd -= ax;
      }
      if (zd>=0) {
	z += sz;
	zd -= ax;
      }
      x += sx;
      yd += ay;
      zd += az;
    }
  }
  else if (ay>=ax && ay>=az) {  // y dominant
    xd = ax-(ay>>1);
    zd = az-(ay>>1);
    for (;;) {
      v = val(x,y,z);
      sum += (y==y1) ? 0.0f : (v+lastv)*0.5f;
      lastv = v;
      float ptau = tau;
//      tau = (dy==0) ? dt*v : sum/(dy*sy)*dt;
      tau = (dy==0) ? dt*v : sum/float((y-y1+sy)*sy)*dt;
      if (y==y2 || tau>=stopAmt){
	if (t){
	  float pcent = (tau>=stopAmt) ? (stopAmt-ptau)/(tau-ptau) : 1.0f;
	  *t = dy==0 ? t1 + pcent*dt :
            t1 + ( (tau>0.0f) ? ( dt * (y-y1-sy+sy*pcent)/(float)dy) : 0.0f );
	}	  
	return tau;
      }
      if (xd>=0) {
	x += sx;
	xd -= ay;
      }
      if (zd>=0) {
	z += sz;
	zd -= ay;
      }
      y += sy;
      xd += ax;
      zd += az;
    }
  }
  else {                      // z dominant
    xd = ax-(az>>1);
    yd = ay-(az>>1);
    for (;;) {
      v = val(x,y,z);
      sum += (z==z1) ? 0.0f : (v+lastv)*0.5f;
      lastv = v;
      float ptau = tau;
//      tau = (dz==0) ? dt*v : sum/(dz*sz)*dt;
      tau = (dz==0) ? dt*v : sum/float((z-z1+sz)*sz)*dt;
      if (z==z2 || tau>=stopAmt){
	if (t){
	  float pcent = (tau>=stopAmt) ? (stopAmt-ptau)/(tau-ptau) : 1.0f;
	  *t = dz==0 ? t1 + pcent*dt :
            t1 + ( (tau>0.0f) ? ( dt * (z-z1-sz+sz*pcent)/(float)dz) : 0.0f );
	}
	return tau;
      }
      if (xd>=0) {
	x += sx;
	xd -= az;
      }
      if (yd>=0) {
	y += sy;
	yd -= az;
      }
      z += sz;
      xd += ax;
      yd += ay;
    }
  }

}

// This is world units IN, and world units OUT.
// So, stopAmt should be in world units (meaning you multiplied the real
// stop tau with gv->units(1) already), and, to get t or the return value
// in meters, you have to multiply with gv->meters(1) afterwards.
float DataSet::normalSum(const SbVec3f &p1_, const SbVec3f &p2_,
			 float *t, const float stopAmt){
  if (stopAmt<=0){
    *t = 0;
    return 0;
  }
  
  // transform vectors into dataset space
  SbVec3f p1, p2;
  float t1, t2;
  clipToVolume(p1_, p2_, p1, p2, t1, t2);
    
  float sum = 0.0;
  if (t1>=t2)
    return sum;

  SbVec3f r = p2-p1;
  float dx_world = (p2_-p1_).length();
  float dx_local =  r.normalize();
  float localToWorld = dx_world/dx_local; // this only depends on transform
                                          // and could be precomputed
  
  // constant
//   float cres = 0.5f*dt;
//   if (t && cres>=stopAmt) *t = t1+dt*stopAmt/cres;
//   return (cres);

  // march
  double prevt = t1;
  double curt = t1 + defaultStepSize*drand48();
  if (curt>t2)
    curt = t2;
  SbVec3f x = p1 + t1*r;
  float newv, oldv = valInterpNoX(x);
  float nexttau, totaltau = 0;
  while (1) {
      
    x = p1 + curt*r;
    float dt = curt-prevt;
    newv = valInterpNoX(x);
    float deltatau = (newv+oldv)*0.5 * dt;
    nexttau = totaltau + deltatau;
    
    if (nexttau > stopAmt){
      float pcent = (stopAmt-totaltau)/deltatau;
      if (t) *t = (prevt + pcent*dt)*localToWorld;
      return stopAmt*localToWorld;
    }
    
    oldv = newv;
    prevt = curt;
    totaltau = nexttau;

    if (curt == t2)
      return totaltau*localToWorld;
    
    curt += defaultStepSize;
    if (curt>t2)
      curt = t2;
  }

}

// create inverse transform matrix
void DataSet::setRotation(SbRotation r){
  SbVec3f center = (max+min)/2.0;
  SbVec3f zero(0,0,0), iden(1,1,1);
  transform.setTransform(zero, r, iden, r, center);
  transform = transform.inverse();
}


void DataSet::clipToVolume(const SbVec3f &p1_, const SbVec3f &p2_,
			   SbVec3f &p1, SbVec3f &p2, float &t1, float &t2){

  // transform vectors into dataset space
  transform.multVecMatrix(p1_, p1);
  transform.multVecMatrix(p2_, p2);
  
  // intersect line with box
  SbVec3f r = p2-p1;
  t1=0.0;
  t2=r.normalize();
  SbVec3f tmin, tmax;
  float eps = 1e-25;  // to prevent divide by zero
  tmin[0] = (min[0]-p1[0])/(r[0]+eps); // x=x0+rt  ->  t=(x-x0)/r
  tmax[0] = (max[0]-p1[0])/(r[0]+eps); 
  tmin[1] = (min[1]-p1[1])/(r[1]+eps); 
  tmax[1] = (max[1]-p1[1])/(r[1]+eps); 
  tmin[2] = (min[2]-p1[2])/(r[2]+eps); 
  tmax[2] = (max[2]-p1[2])/(r[2]+eps);

  // swap min/max
  float tmp;
  if (r[0]<0.0){ tmp = tmin[0]; tmin[0] = tmax[0]; tmax[0] = tmp; }
  if (r[1]<0.0){ tmp = tmin[1]; tmin[1] = tmax[1]; tmax[1] = tmp; }
  if (r[2]<0.0){ tmp = tmin[2]; tmin[2] = tmax[2]; tmax[2] = tmp; }

  float mint = MAX3(tmin[0], tmin[1], tmin[2]);
  float maxt = MIN3(tmax[0], tmax[1], tmax[2]);

  // clip p1 and p2 to box
  if (mint>t1) t1 = mint;
  if (maxt<t2) t2 = maxt;

  return;
}


void DataSet::writeData(const string &fileName){

  FILE *fp = fopen(fileName.c_str(), "wb");
  if (fp == NULL){
    cerr << "Error: Could not open file: " << fileName
	 << " for writing: " << strerror(errno) << "\n";
    return;
  }

  for (int k=0; k<dim[Z]; k++){
    for (int j=0; j<dim[Y]; j++){
      for (int i=0; i<dim[X]; i++){
	float f = val(i,j,k);
	fwrite(&f, sizeof(float), 1, fp);
      }
    }
  }

  fclose(fp);
  cerr << "Wrote " << dim[0]*dim[1]*dim[2]*sizeof(float)
       << " bytes to " << fileName << ".\n";
    
}

// color storage
void DataSet::initColors(float mpu){

  // calculate color table
  colorsTable = new SbColor[colorSamples];
  for (int i=0; i<colorSamples; i++){
    float temp = minTemp+(i/(float)(colorSamples-1))*(maxTemp-minTemp);
    colorsTable[i] = Spectrum(temp).getRGB()*(1.0f/(4.0f*M_PI));
  }
  colorsTable[0].setValue(0,0,0); // make coldest/darkest one black

  // calculate color volume
  colors = new int[sdim[0]*sdim[1]*sdim[2]];
  cerr << "Precomputing blackbody spectral emission...\n";
  for (int k=0; k<dim[Z]; k++){
    for (int j=0; j<dim[Y]; j++)
      for (int i=0; i<dim[X]; i++){
	const int q = Fx[i] + Fy[j] +Fz[k];
	float temp = data[q];
	if (temp>minTemp)
	  colors[q] = colorIdxLookup(temp);
	  //Spectrum(data[q]).getRGB()*(1.0f/(4.0f*M_PI));
	else
	  colors[q] = 0; // black
	  //colors[q].setValue(0,0,0);
      }
    fprintf(stderr, "\r%02.2f%%", k/(float)dim[Z]*100.0f);
  }
  cerr << flush << "\rDone.                         " << endl;
}

SbColor DataSet::colorVal(const SbVec3f &x_) const {

  // constant
  //return (0.5);
  
  SbVec3f x;
  transform.multVecMatrix(x_, x);
  
  // convert position into index+offset
  int i, j, k;
  float a, b, c;
  xToijkabc(x, i, j, k, a, b, c);

  int i1 = i+1;
  int j1 = j+1;
  int k1 = k+1;
  
  // interpolate
  SbColor f =
    (1.0f-a)*(1.0f-b)*(1.0f-c) * colorVal( i,  j,  k) +
    (1.0f-a)*(1.0f-b)*(c)      * colorVal( i,  j, k1) +
    (1.0f-a)*(b)*(1.0f-c)      * colorVal( i, j1,  k) +
    (1.0f-a)*(b)*(c)           * colorVal( i, j1, k1) +
    
    (a)*(1.0f-b)*(1.0f-c)      * colorVal(i1,  j,  k) +
    (a)*(1.0f-b)*(c)           * colorVal(i1,  j, k1) +
    (a)*(b)*(1.0f-c)           * colorVal(i1, j1,  k) +
    (a)*(b)*(c)                * colorVal(i1, j1, k1);

  //static Noise h;
  //f *= h.val(x/5.0f, 0.5, .85);

  return(f);
}

//g++ DataSet.cxx -I/usr/X11R6/include -I/usr/include/OpenEXR -lInventor -O2 -DDATASET_MAIN -o dataSetTest
#ifdef DATASET_MAIN
  SoSeparator *scnRoot;
int main(int argc, char *argv[]){
#else
int DataSetMain(int argc, char *argv[]){
#endif

  DataSet ds;
  int dim[] = { 20, 20, 20 };
  ds.useData("", dim, 0, 0, SbVec3f(0,0,0), SbVec3f(100,100,100));
  ds.writeData("test20x20x20.dat");

  return 0;
}
