//WWWWWW'''''''WW''''''''''''''''WWW'''''''WWWWWW'''''''''''''''''''''''''WWW'
//  WW                            WW         WW                            WW 
//  WW    W WW WW  WWW WWW   WWW  WW   WW    WW    W WW  WW   WW      WWW  WW 
//  WW    WW   WW W WW WW W W WW  WW  WWWW   WW    WW   WWWW WWWW    WW W  WW 
//  WW    WW   WW W WW WW W W WW  WW  WW     WW    WW   WW   WW      WW    WW 
//  WW    WW   WW  WWW WW W  WWW WWWW  WWW   WW    WW    WWW  WWW     WWW WWWW
//                            WW                                              
//''''''''''''''''''''''''''WWW'''''''''''''''''''''''''''''''''''''''''''''''

#ifndef OCTREECLASS
#define OCTREECLASS

#include "Primitive.h"
#include "RayAccel.h"

class Ray; // forward declaration of Ray class
class BBox;

#define EPSILON_OS  0.03    // (Octree Size)
#define EPSILON_OI  0.00005 // (Octree Include)
//#define SHOWTRAV 1


class Octree : public RayAccel {
private:

  // cell
  typedef struct node_st {
    struct node_st *child[8];   // 32 bytes
    double mid[3];              // 24 bytes
    PrimList *primList;

#ifdef SHOWTRAV  
    // debugging
    SbVec3f min, max;
    int depth;
#endif
  } node;
  
  // parameters struct
  typedef struct treeParam {
    Ray *r_p;                 // ray
    SbVec3f invDir;
    double t0, t1;
    unsigned char a;   // used for octree traversal algorithm
  } treeParam;
  
  const static int delta[8][3];
  int octreeDebug;
  int maxItems;
  int maxDepth;
  node *root;
  node *compactNodes;
  int compactNodesSize;
  PrimList primList;
  unsigned int rayId;
  pthread_mutex_t lock;

  // temporary!
  int rays;

  // also keep bounding sphere of specular primitives
  SbVec3f shiny_min, shiny_max;

  
  void printStats(node *ptr, int *numCubes = NULL, int *numPrims = NULL);
  double findDividingPlane(node *n, int axis, const SbVec3f &vmin, const SbVec3f &vmax,
                           const vector<BBox> &bboxs);
  void refineNode(node *n, int depth, int simpleDiv,
		  const SbVec3f &vmin, const SbVec3f &vmax);
  inline int new_node(double &t1, int i1, double &t2,
		      int i2, double &t3, int i3) const;
  int first_node(double &tx0, double &ty0, double &tz0,
		 double &txm, double &tym, double &tzm) const;
  bool proc_subtree(double &tx0, double &ty0, double &tz0,
                    double &tx1, double &ty1, double &tz1,
                    node *n, treeParam &tp) const;
  bool ss_intersectRay(Ray &r, double t0=1e-5, double t1=1e8);
  void deleteNode(node *n);

public:

  Octree();
  ~Octree();

  // Primitive methods
  PrimType getType() const { return PRIMTYPE_OCTREE; }
  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8);
  SbVec3f getMinShiny() const { return shiny_min; }
  SbVec3f getMaxShiny() const { return shiny_max; }
  
  void reset();

  int add(Primitive *t){

    // add primitives
    numPrims++;
    primList.add(t);    
  }
  void reserve(int n) { primList.reserve(n); }

  void printStats();
  void setParams(int maxD, int maxI, int debug);
  void refine(int simpleDiv = 0);
  void reorderNodes();
  
};

#endif /* OCTREECLASS */
