#include "Plane.h"
#include "Main.h"
#include "debug.h"

Material Plane::defaultMat;


DMPlane::DMPlane(float vertFOV, int vertRes) : Plane() {
  maxHeight = 1;
  errorAt1 = tanf(vertFOV*M_PI/180. / vertRes);

//   noise.octaves = 10;
//   noise.amplitude = 1;
//   float scale = 3;
//   noise.frequency = SbVec3f(scale,scale,scale);
//   //noise.translation = -offset;
//   //noise.bias = clamp(bias,0,1);
//   //noise.gain = clamp(gain,0,1);
//   noise.ampJump = 1;
//   noise.freqJump = 2;
//   noise.sign=false;

  noise.octaves = 8;
  noise.amplitude = .99/1.92651;
  float scale = 2.638*.173;
  noise.frequency = SbVec3f(scale,scale,scale);
  noise.bias = .4257;
  noise.gain = .8769;
  noise.translation = -SbVec3f(.3742, .135, -.5825);
  noise.freqJump = 2.029;
  noise.ampJump = 1;
  noise.sign=true;
  
  noise.noise(SbVec3f(0,0,0)); // initialize exponent array
}

// This code is slow as hell now
bool DMPlane::intersectRay(Ray &r, double t0, double t1) {
  r.t = -1;
  r.prim = NULL;
  r.bc = SbVec3f(1,0,0);

  //if (r.ignorePrim == this) return false;

  float h = maxHeight;
  SbVec3f p = onb.toONBSpace(r.pos);
  SbVec3f d = onb.toONBSpace(r.dir);

  const double myEps = 0.001; // tanscale(fov_h/res_h)   // error at 1u
  const double stept = myEps;//1./100;//2048.;//1/256.;//0.01;
  const double maxGrad = 2;
  const double maxGradInv = 1/maxGrad;

  double h0, h1, hm;
  double d0, d1, dm;
  Vec3d p0, p1, pm; // in ONB space


  // intersect ray with slab
  float posDotN = p[2]; //p.dot(onb.getW());
  float dirDotN = d[2]; //d.dot(onb.getW());

  // ray parallel and outside slab?
  static const float min = 0.00001;//FLT_MIN*32;
  bool parallel = fabs(dirDotN) < min;
  if (parallel &&
      (posDotN<dist || posDotN>(dist+h)) )
    return false;

  // clamp t0 t1 to volume  
  if (!parallel){
    double m0 = -(posDotN-dist)/dirDotN;
    double m1 = -(posDotN-(dist+h))/dirDotN;

    // order m0, m1
    if (m1<m0){
      double t = m1;
      m1 = m0;
      m0 = t;
    }

    t0 = MAX2(t0, m0);
    t1 = MIN2(t1, m1);
  }
  
  // more near and far clipping
  float nearClip = 0.01; //0.00001;
  float farClip = 10000;
  t0 = MAX2(t0, nearClip);
  t1 = MIN2(t1, farClip);
  
  if (t1<=t0)
    return false;
  
  d0 = t0;
  
  p0 = p+d*d0;
  h0 = height(p0, d0);
  
  // check for already near

//   if (fabs(p[2]-h0)<myEps) {

//     if (DEBUG_STATE)
//       cerr << "hit!\n";
      
//     // hit heightfield
//     r.t = d0;
//     r.prim = this;
//     r.bc = p0;
//     return true;
//   }
  
  bool above = p0[2]>h0;

  if (DEBUG_STATE)
    cerr << "Begin marching\n";

  // iterate
  d1 = d0;
  p1 = p0;
  h1 = h0;
  unsigned int steps = 0;
  while (d1<t1) {
    
    if (DEBUG_STATE)
      cerr << __LINE__ << ": d1=" << d1 << "\n";

    d0 = d1;    // distance
    p0 = p1;    // position (u,v, height)
    h0 = h1;    // heightfield height
    
    d1 += MAX2(d1*myEps, fabs((p1[2]-h1)*maxGradInv));           //stept;
    steps++;
    
    if (d1>t1){
      if (DEBUG_STATE)
        cerr << __LINE__ << ": Reached end\n";
      d1 = t1;
    }

    if (DEBUG_STATE)
      cerr << __LINE__ << ": d1=" << d1 << "\n";
    
    // get height
    if (DEBUG_STATE)
      cerr << __LINE__ << ": p1= " << p1 << "\n";
    
    p1 = p+d*d1;
    h1 = height(p1, d1);
    if (DEBUG_STATE)
      cerr << __LINE__ << ": h1=" << h1 << "\n";
    
    // detect crossing
    if (above && p1[2]<=h1 ||
        !above && p1[2]>=h1 /*||
        fabs(p1[2]-h1)<myEps*/) {


      if (DEBUG_STATE)
        cerr << "hit... now refine...\n";

      // refine intersection
      double diff;
      do {
        dm = (d0+d1)*0.5;
        diff = d1-dm;
        pm = p+d*dm;
        hm = height(pm, dm);

        if (above && pm[2]<hm ||
            !above && pm[2]>hm /*||
            fabs(pm[2]-hm)<myEps*/) {
          d1=dm;
          p1=pm;
        } else {
          d0=dm;
          p0=pm;
        }
          
      } while (diff>.01);
        
      // hit heightfield
      r.t = d0; // choose nearer hit, give next ray a better
                // chance of not self-intersecting
      r.prim = this;
      r.bc = p0;
      //r.bc[2] = steps;
      if (DEBUG_STATE)
        cerr << "returning t=" << d0 << "\n";
      return true;
    }

  }

  if (DEBUG_STATE)
    cerr << __LINE__ << ": no hit\n";
  return false;
}

SbVec3f DMPlane::getNormal(const SbVec3f &bc) const {

  const float du = 0.001;
  const float duInv = 1/du;
  const float dv = du;
  const float dvInv = duInv;

  float h = maxHeight;

  SbVec3f p0 = bc; 
  SbVec3f pu = bc; pu[0]+=du; // + onb.getU()*du;
  SbVec3f pv = bc; pv[1]+=dv; //+ onb.getV()*dv;
  
  float h0 = height(Vec3d(p0[0], p0[1], 1.), 1);
  float hu = height(Vec3d(pu[0], pu[1], 1.), 1);
  float hv = height(Vec3d(pv[0], pv[1], 1.), 1);

  SbVec3f norm((hu-h0)*duInv, (hv-h0)*dvInv, -1);
  norm = onb.toWorldSpace(norm);
  norm.normalize();
  return norm;
}

SbColor DMPlane::getIntDiffColor(const SbVec3f &bc) const {
  return SbColor(.5,.5,.5);
//   float v = fmodf(bc[2],10000.)/10000.;
//   SbColor c(v, 0, 0);
//   //cerr << "bc=" << bc << " c=" << c << "\n";
//   return c;
}
