#ifndef PLANE_H
#define PLANE_H

#include <iostream>
#include <iomanip>
#include "Primitive.h"
#include "Material.h"
#include "Misc.h"
#include "Noise.h"
#include "debug.h"
#include "Main.h"

using namespace std;

class Plane : public Primitive {

protected:

  static Material defaultMat;
  
  SbVec3f norm;
  float dist;
  Material *mat;
  ONB onb;
  
public:

  Plane() {
    norm = SbVec3f(0,1,0);
    dist = 0;
    mat = &defaultMat;
    onb.constructFromW(norm);
  }
  
  virtual PrimType getType() const { return PRIMTYPE_PLANE; }

  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8) {
    static const float min = 0.00001;//FLT_MIN*32;
    
    r.t = -1;
    r.prim = NULL;
    r.bc = SbVec3f(1,0,0);

    if (r.ignorePrim == this) return false;
    
    float dirDotN = r.dir.dot(norm);
    if (fabs(dirDotN) < min)
      return false;

    float posDotN = r.pos.dot(norm);
    double t = -(posDotN-dist)/dirDotN;

    if (t<t0 || t>t1)
      return false;

    r.t = t;
    r.prim = this;
    return true;
  }

  SbVec3f getMin() const {
    // temporary until I fix pane->rayAccel to be Primitive*
    return SbVec3f(-1000, 0, -1000);
    
    float v = FLT_MAX;
    return SbVec3f(v,v,v);
  }
  SbVec3f getMax() const {
    // temporary until I fix pane->rayAccel to be Primitive*
    return SbVec3f(1000, 0, 1000);
    
    float v = -FLT_MAX;
    return SbVec3f(v,v,v);
  }

  Material* getMat() { return mat; }
  const Material* getMat() const { return mat; }
  SbVec3f getNormal(const SbVec3f &bc)  const { return norm; }

  static Material* getDefaultMat() { return &defaultMat; }

};

class DMPlane : public Plane {

private:
  float maxHeight;
  float errorAt1;  // allowable error at d=1 (needed for octaves)
                   // = tan(fov/res)
  mutable Noise noise;

public:
  DMPlane(float vertFOV=45, int vertRes=768);
  
  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8);

  SbVec3f getMin() const {
    // temporary until I fix pane->rayAccel to be Primitive*
    return SbVec3f(-1000, 0, -1000);
    
    float v = FLT_MAX;
    return SbVec3f(v,v,v);
  }
  SbVec3f getMax() const {
    // temporary until I fix pane->rayAccel to be Primitive*
    return SbVec3f(1000, maxHeight, 1000);
    
    float v = -FLT_MAX;
    return SbVec3f(v,v,v);
  }

  SbVec3f getNormal(const SbVec3f &bc) const;

  // testing purposes
  SbColor getIntDiffColor(const SbVec3f &bc) const;

  double height(const Vec3d &x) const {
    return fabs(noise.noise(x[0], x[1], 1.)*maxHeight);
  }
  float height(const Vec3d &x, float d) const {
    return height(x); // don't use adaptive octaves for now

    // clamp d
    if (d<1) d = 1;

//    static int minOct = 30;
    // from QAEB paper by K. Musgrave, bottom of page 2
//    cerr << __LINE__ << " wtf errorAt1*d=" << errorAt1*d << "\n";
//    int oct = (int)( 0.5 + -3.5-fastlog2(float(errorAt1*d)) );
//     int oct = (int)( 0.5 + -3.5 - log(errorAt1*d)*(1/log(2)) );
//    cerr << __LINE__ << " wtf oct= " << oct << "\n";

//     if (oct<1)
//       oct = 1;
    
//     if (oct>10){
//       cerr << "oct=" << noise.octaves << " d=" << d << " errorAt1=" << errorAt1 <<"\n";
//       exit(1);
//     }

//     if (oct<minOct){
//       minOct = oct;
//       cerr << "minOct = " << minOct << "\n";
//     }

//    noise.octaves = oct;
    return height(x);
  }
  
};


#endif
