#include "Main.h"
#include "Primitive.h"
#include "debug.h"

const SbColor Primitive::Black(0,0,0);
const SbColor Primitive::White(1,1,1);
const SbVec3f Primitive::XDir(1,0,0);
const SbVec3f Primitive::YDir(0,1,0);
const SbVec3f Primitive::ZDir(0,0,1);
const PrimDerivs Primitive::DefaultDerivs(XDir, YDir,
                                          SbVec3f(0,0,0), SbVec3f(0,0,0));
Material Primitive::defaultMat;

// bool Primitive::intersectCube(const SbVec3f &min, const SbVec3f &max) const {

//   SbVec3f myMin = getMin();
//   SbVec3f myMax = getMax();

//   return
//     myMin[0] < max[0] && myMax[0] > min[0] &&
//     myMin[1] < max[1] && myMax[1] > min[1] &&
//     myMin[2] < max[2] && myMax[2] > min[2];
// }

bool Primitive::intersectCube(const double min[3], const double max[3]) const {

  SbVec3f myMin = getMin();
  SbVec3f myMax = getMax();

  return
    myMin[0] <= max[0] && myMax[0] >= min[0] &&
    myMin[1] <= max[1] && myMax[1] >= min[1] &&
    myMin[2] <= max[2] && myMax[2] >= min[2];
}

bool Instance::intersectRay(Ray &r, double t0, double t1){
  SbVec3f post, dirt;
  mi.multVecMatrix(r.pos, post);
  mi.multDirMatrix(r.dir, dirt);
  dirt.normalize();
  
  Ray rt(post, dirt, r, r.prim);

  if (prim->intersectRay(rt, t0, t1)){

    SbVec3f p;
    m.multVecMatrix(rt.point(), p);
      
    // hit
    r.t = (r.pos-p).length();
    r.prim = this;
    r.bc = rt.bc;
    return true;
  }

  // no hit
  r.t  = -1;
  r.prim = NULL;
  r.bc = SbVec3f(1,0,0);
  return false;
}
