
#ifndef SKYCLASS
#define SKYCLASS

#include "Main.h"
#include "GlobalVars.h"
#include "Ray.h"
#include "Pane.h"
#include "EnvironMap.h"

class Sky : public EnvironMap {

  SbVec3f sunBasisU, sunBasisV, sunBasisW;  // W points toward sun
  float sunRadius, sunRadiusAngle, sunDist, sunOmega; // (omega is solid angle)
  float turbidity, thetaSolar, thetaSolarSafe;
  SbColor sunRadiance;
  
  private:

  float AY,BY,CY,DY,EY,Ax,Bx,Cx,Dx,Ex,Ay,By,Cy,Dy,Ey;
  float Yz,xz,yz;
  EnvironMap sky;  // don't rely exclusively on subclass because virtual method calls are
                   // tricky, and we want different rotation matrices anyway
         
  
  public:
  
  Sky();
  ~Sky();
  
  virtual void init(GlobalVars *gv) { init(gv, 17.0); }
  virtual void init(GlobalVars *gv, float time);
  virtual SbColor impSample(Ray &r, SbVec3f &n, SbVec3f &ng, int simple=0);
  //virtual SbColor volImpSample(Ray &r, int simple);  
  virtual float getSample(double eps1, double eps2,
		  SbVec3f &dir, SbColor &c);
  virtual SbColor getRadiance(SbVec3f &dir) { return getRadiance(dir, true); }
  virtual SbColor getRadiance(SbVec3f &dir, bool includeSun);
  virtual SbColor getEmissiveFlux(float receiverArea);
  
};

#endif /* ENVIRONMAPCLASS */
  
