//WWWWWW'''''''''''''WW'''''''''''''''''''''''''WWW''''''''''''''''''
//  WW               WW                          WW                  
//  WW     WW  WW W WWWW W WW W WW  WW      WWW  WW   WWW  WWWW  WWWW
//  WW    WWWW  WW   WW  W WW WW   WWWW    WW W  WW  W WW WWW   WWW  
//  WW    WW    WW   WW  W WW WW   WW      WW    WW  W WW   WWW   WWW
//  WW     WWW W WW  WW   WWW WW    WWW     WWW WWWW  WWW WWWW  WWWW 
//                                                                   
//'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

#include "Texture.h"
#include "debug.h"
#include "Main.h"

Texture::Texture(){
  wrapS = SoTexture2::REPEAT;
  wrapT = SoTexture2::REPEAT;
  model = SoTexture2::MODULATE;
  numComps = 3;
  bumpMap = 0;
  bumpScale = 1.0;
  img = NULL;
  imgTexture = NULL;
  size = SbVec2s(0, 0);
  blend = SbColor(0.0, 0.0, 0.0);
  stex[0] = ttex[0] = 0.0;
  stex[1] = 1.0; ttex[1] = 0.0;
  stex[2] = 0.5; ttex[2] = 1.0;
}

Texture::~Texture(){
  if (texture_maps.count(img) != 0){
    delete [] texture_maps[img];
    texture_maps.erase(img);
  }
}

Texture::Texture(int s, int t, int m,
		 SbColor b, int n,  SbVec2s &siz,
		 const unsigned char *_img,
		 int _bumpMap, float _bumpScale){
  wrapS = s;
  wrapT = t;
  model = m;
  blend = b;
  numComps = n;
  size = siz;
  img = _img;
  bumpMap = _bumpMap;
  bumpScale = _bumpScale;
  addTextureMap(img, numComps, size, imgTexture);

  // stex[], ttex[] get filled in by Triangle(...)
}


SbColor Texture::getColor(const unsigned char *img, int numComps,
			  float s, float t,
			  SbColor &color) const {

  float r, g, b;
  float i_f, j_f;
  float sp, tp;
  int i, j, i1, j1;

//  cerr << "s=" << s << " t=" << t << "\n";
  // do wrapping
  if (wrapS == SoTexture2::REPEAT){
    s -= (float)((int)s);
    if (s<0.0)
      s += 1.0;

    t -= (float)((int)t);
    if (t<0.0)
      t += 1.0;
  }
  // or clamp
  else {
    s = clamp(s, 0.0f, 1.0f);
    t = clamp(t, 0.0f, 1.0f);
  }
  
  // calculate indices
  i  = (int)floor(s*size[0]);
  j  = (int)floor(t*size[1]);
  i1 = i+1;
  j1 = j+1;

  if (wrapS == SoTexture2::REPEAT){
    if (i1>size[0]-1)
      i1 = 0;
    if (j1>size[1]-1)
      j1 = 0;
  }
  else {
    i1 = clamp(i1, 0, size[0]-1);
    j1 = clamp(j1, 0, size[1]-1);
  }
  
//  i = clamp(i, 0, size[0]-1);
//  j = clamp(j, 0, size[1]-1);
 
  // calculate texel ratios
  sp = s*size[0]-i;
  tp = t*size[1]-j;
	 
  // perform hermite smoothing
  sp = 3.0*sp*sp-2.0*sp*sp*sp;
  tp = 3.0*tp*tp-2.0*tp*tp*tp;

  if (numComps == 3){
	 
    r = (1.0-sp)*(1.0-tp)*img[size[0]*3*j +i *3+R] +
              sp*(1.0-tp)*img[size[0]*3*j +i1*3+R] +
        (1.0-sp)*tp      *img[size[0]*3*j1+i *3+R] +
              sp*tp      *img[size[0]*3*j1+i1*3+R];

    g = (1.0-sp)*(1.0-tp)*img[size[0]*3*j +i *3+G] +
             sp *(1.0-tp)*img[size[0]*3*j +i1*3+G] +
        (1.0-sp)*tp      *img[size[0]*3*j1+i *3+G] +
              sp*tp      *img[size[0]*3*j1+i1*3+G];

    b = (1.0-sp)*(1.0-tp)*img[size[0]*3*j +i *3+B] +
              sp*(1.0-tp)*img[size[0]*3*j +i1*3+B] +
        (1.0-sp)*tp      *img[size[0]*3*j1+i *3+B] +
              sp*tp      *img[size[0]*3*j1+i1*3+B];

    color = SbColor(r/255.0, g/255.0, b/255.0);
  }
  else if (numComps == 1){

    // debugging
    //cerr << "numComps = 1!!! \n";
    
    r = (1.0-sp)*(1.0-tp)*img[size[0]*j +i ] +
      sp*(1.0-tp)      *img[size[0]*j +i1] +
      (1.0-sp)*tp      *img[size[0]*j1+i ] +
      sp*tp            *img[size[0]*j1+i1];
      
    color = SbColor(r/255.0, r/255.0, r/255.0);
  }

  return color;
}


// texture filtering from "Realistic Ray Tracing" by Peter Shirley p.61
SbColor Texture::getColor(float s, float t, const SbColor &diffuse) const {
  SbColor color;
  getColor(imgTexture, numComps, s, t, color);

  // blend model
  // multiply components by interpolated diffuse color
  if (model == SoTexture2::MODULATE){

    // this uses the same diffuse color barycentric interpolation as in
    // Triangle::getIntDiffColor()
    vecScaleItself(color, diffuse);
  }

  // use texture as blending factor between diffuseColor and blend color
  // white texture = 100% blend color, black = 100% diffuseColor
  else if (model == SoTexture2::BLEND){

    float intensity = vecAvg(color);
    color = intensity*blend + (1.0f-intensity)*diffuse;
  }
    
  return color;
}

SbColor Texture::getColor(float s, float t) const {
  SbColor color;
  getColor(imgTexture, numComps, s, t, color);
  return color;
}

typedef map<const unsigned char *, unsigned char *> PointerMap;
PointerMap Texture::texture_maps;

void Texture::addTextureMap(const unsigned char *img,
			    int numComps, SbVec2s size,
			    unsigned char *&imgTexture){

  // already copied map?
  if (texture_maps.count(img)){
    imgTexture = texture_maps[img];
    return;
  }
  
  // create new map
  int bytes = size[0]*size[1]*numComps;
  unsigned char *img2 = new unsigned char[bytes];
  if (img2 == NULL){
    cerr << "Error: Could not allocate memory ("
         << bytes << " bytes) for texture map!\n";
    exit(1);
  }
      
  // Copy and correct (degamma) map
  for (int i=0; i<bytes; i++){
    float x = (float)img[i] * 1./255.;   // convert to float
    float r = powf(x, 2.2);              // degamma image
    r = clamp(r, 0.0f, 1.0f);
    img2[i] = (unsigned char)(r*255.0);
  }
  
  imgTexture = texture_maps[img] = img2;
}


// returns dh/du, dh/dv, h
SbVec3f Texture::getDeriv(float u, float v, float du, float dv) const {

//   if (!bumpMap)
//     return SbVec3f(0, 1, 0);

  if (!du)
    du = 0.5f/size[0];
  if (!dv)
    dv = 0.5f/size[1];
  
  SbColor color0, color1, color2;
  SbColor color1_, color2_;
  getColor(imgTexture, 3, u,    v,    color0);
  getColor(imgTexture, 3, u+du, v,    color1);
  getColor(imgTexture, 3, u-du, v,    color1_);
  getColor(imgTexture, 3, u,    v+dv, color2);
  getColor(imgTexture, 3, u,    v-dv, color2_);

  float b0, b1, b2;
  float b1_, b2_;
  b0 = brightness(color0);
  b1 = brightness(color1);
  b2 = brightness(color2);
  b1_ = brightness(color1_);
  b2_ = brightness(color2_);

  // central difference seems to look better
  // than just forward difference
  SbVec3f deriv;
  deriv[R] = (b1-b1_)*0.5*size[0]; // size[0]=1/du;
  deriv[G] = (b2-b2_)*0.5*size[1]; // size[1]=1/dv;
  deriv[B] = b0;
  
  return deriv;
}
  
