#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <zlib.h>
#include <math.h>
#include <iomanip>

#include <Inventor/SbLinear.h>
#include <Inventor/SoDB.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoOutput.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/actions/SoToVRMLAction.h>

using namespace std;

void printUsage(char **argv){
  cerr << "Usage: " << argv[0] << " scene.wrl scene.iv\n";
  exit(1);
}

ostream& operator<< (ostream& os, const SbVec3f& x){
  os << setiosflags(ios::showpoint | ios::fixed) << setprecision(8);
  os << "(" << x[0] << ", " << x[1] << ", " << x[2] << ") ";
}

int main(int argc, char *argv[]){

  SoDB::init();
  
  int c;
  while (1){
    c = getopt(argc, argv, "");

    if (c == -1)
      break;

    switch (c){
    default:
      cerr << "Unknown option found.\n";
    }
  }

  if (optind != argc-2){
    printUsage(argv);
  }

  char *file1 = argv[optind++];
  char *file2 = argv[optind];

  // read scene data
  SoInput *sceneIn = new SoInput();
  if (!sceneIn->openFile(file1)){
    cerr << "Unable to open scene file: " << file1 << endl;
    exit(1);
  }

  cerr << "Reading scene file" << file1 << "\n";

  // Read scene
  SoSeparator *scene = SoDB::readAll(sceneIn);
  scene->ref();
  if (!scene){
    cerr << "Error: scene separator is null!" << endl;
    exit(1);
  }
  delete sceneIn;

  // Convert to VRML1.0
  SoToVRMLAction ca;
  ca.apply(scene);
  SoSeparator *scene2 = (SoSeparator *)ca.getVRMLSceneGraph();

  SoOutput *out = new SoOutput();
  out->openFile(file2);
  SoWriteAction *wa = new SoWriteAction(out);
  wa->apply(scene2);
  out->closeFile();
  cerr << "Wrote scene (VRML1.0) to " << file2 << ".\n";
  delete out;
  delete wa;

  return(0);
}
