#include <iostream>
#include <stdlib.h>
#include <sys/time.h>

using namespace std;

// Compute the delay between t1 and t2 in milliseconds
inline long delay (struct timeval t1, struct timeval t2){
  long d;
    
  d = (t2.tv_sec - t1.tv_sec) * 1000;
  d+= ((t2.tv_usec - t1.tv_usec + 500) / 1000);
  return(d);
}

// time a set of instructions
#define TIME(x) { struct timeval start, end;  \
                gettimeofday(&start, NULL); \
                x;                          \
 	        gettimeofday(&end, NULL);   \
	        cerr << "Time: "  << delay(start, end) << "ms" << endl; }

void bubbleSort(float *vals, int n){
  for (int i=n-1; i>=0; i--)
    for (int j=0; j<i; j++)
      if (vals[j]>vals[j+1]){
	// swap
	const float tmp = vals[j];
	vals[j]=vals[j+1];
	vals[j+1] = tmp;
      }
}

void printVals(float *vals, int numVals){
  cerr << "Array:\n";
  for (int i=0; i<numVals; i++) cerr << vals[i] << "\n";
}


// flip a float for sorting
// finds sign of floating point number:
//   1 (negative): flips all bits
//   0 (positive): flips sign
inline unsigned int FloatFlip(unsigned int f){
  unsigned int mask = -(int)(f>>31) | 0x80000000;
  return f ^ mask;
}
inline unsigned int FloatFlipF(unsigned int &f){
  unsigned int mask = -(int)(f>>31) | 0x80000000;
  f ^= mask;
}

// flips a float back
inline unsigned int FloatFlipI(unsigned int f){
  unsigned int mask = ((f>>31)-1) | 0x80000000;
  return f ^ mask;
}

#define _0(x) (x & 0x7FF)
#define _1(x) (x >> 11 & 0x7FF)
#define _2(x) (x >> 22 )

// radix sort by Michael Herf
// http://www.stereopsis.com/radix.html
// email to use: stereopsis att mherf dt com 
void radixSort(float *vals, int numVals){
  unsigned int i;
  unsigned int *sort = new (unsigned int)[numVals];
  unsigned int *orig = (unsigned int *)vals;

  const unsigned int kHist = 2048;
  unsigned int b0[kHist * 3];
  unsigned int *b1 = b0+kHist;
  unsigned int *b2 = b1+kHist;

  for (i=0; i<kHist*3; i++)
    b0[i] = 0;

  // 1) parallel histogramming pass
  for (i=0; i<numVals; i++){
    unsigned int f = FloatFlip(orig[i]);
    b0[_0(f)]++;
    b1[_1(f)]++;
    b2[_2(f)]++;
  }

  // 2) sum histograms: each histogram entry records the
  // number of values preceding itself.
  unsigned int sum0=0, sum1=0, sum2=0, tsum;
  for (i=0; i<kHist; i++){
    tsum = b0[i]+sum0;
    b0[i] = sum0-1;
    sum0 = tsum;

    tsum = b1[i]+sum1;
    b1[i] = sum1-1;
    sum1 = tsum;

    tsum = b2[i]+sum2;
    b2[i] = sum2-1;
    sum2 = tsum;
  }

  // byte 0
  // floatflip entire value, read/write hist, write out flipped
  for (i=0; i<numVals; i++){
    unsigned int f = orig[i];
    FloatFlipF(f);
    unsigned int pos = _0(f);
    sort[++b0[pos]] = f;
  }

  // byte 1
  // read/write hist, copy
  for (i=0; i<numVals; i++){
    unsigned int s = sort[i];
    unsigned int pos = _1(s);
    orig[++b1[pos]] = s;
  }

  // byte 2
  // read/write hist, copy & flip out
  for (int i=0; i<numVals; i++){
    unsigned int s = orig[i];
    unsigned int pos = _2(s);
    sort[++b2[pos]] = FloatFlipI(s);
  }

  // copy
  for (i=0; i<numVals; i++)
    orig[i] = sort[i];
  delete sort;
}

int main(int argc){

  const int numVals = 1<<16;
  cerr << "Vals="<< numVals << "\n";
  float vals[numVals];
  srand48(0);
  for (int i=0; i<numVals; i++)
    vals[i] = drand48()*4096-2048;

  // sort vals
  //printVals(vals, numVals);
  //TIME(bubbleSort(vals, numVals));
  TIME(radixSort(vals, numVals));
  //printVals(vals, numVals);

  // check vals
  bool valid=true;
  for (int i=0; i<numVals-1; i++)
    if (vals[i]>vals[i+1]){
      valid=false;
      cerr << "Error at val="<<i<<": "
	   << vals[i] << " " << vals[i+1] << "\n";
    }
  if (valid) cerr << "Array is sorted in ascending order.\n";
  else cerr << "ERROR: Array is NOT sorted!\n";
  
  return 0;
}
