#include <iostream>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <zlib.h>
#include <math.h>
#include <iomanip>
#include <errno.h>

#include <Inventor/SoInput.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoOffscreenRenderer.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/actions/SoGLRenderAction.h>

#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoResetTransform.h>
#include <Inventor/SoInput.h>

using namespace std;

// variables
bool useSmoothing = FALSE;
bool renderQuit = FALSE;
int numSamples = 1;
int imgDim[2] = { 1024, 768 };
char *outFile = "image.png";
char *inFile = "scene.iv";
char *cameraFile = NULL;
SoSeparator *scene = NULL;
SoXtExaminerViewer *myViewer;


// This code is mainly from SGI Open Inventor (and Pane)
// inventor/libSoXt/src/viewers/SoXtViewer.c++
////////////////////////////////////////////////////////////////////////
//
// Description:
//    adds a directional light to the scene graph that is a headlight
//    positioned over the left shoulder of the camera.
//
void setCameraAndHeadlight(SoSeparator *root, char *readCamFile = NULL)
//
////////////////////////////////////////////////////////////////////////
{

    
  //
  // find the camera parent to insert/remove the headlight
  //
  // Does scene have a camera?
  SoCamera *newCam = NULL;

  if (readCamFile){

    cerr << "Reading camera from " << readCamFile << "\n";
    
    SoInput *sceneIn = new SoInput();
    if (!sceneIn->openFile(readCamFile)){
      cerr << "Unable to open scene file: " << readCamFile << endl;
      exit(1);
    }

    SoSeparator *camScene = SoDB::readAll(sceneIn);
    if (!camScene){
      cerr << "Error: Camera scene separator is null!" << endl;
      exit(1);
    }
    camScene->ref();

    SoSearchAction sa;
    sa.setType(SoCamera::getClassTypeId());
    sa.setSearchingAll(FALSE); // don't look under off switches
    sa.apply(camScene);
    if (sa.getPath())
      newCam = (SoCamera *)((SoFullPath *)sa.getPath())->getTail();
    else {
      cerr << "Could not find camera!\n";
      exit(1);
    }

    // remove other cameras.. ?
    sa.setSearchingAll(FALSE); // don't look under off switches
    sa.setFind(SoSearchAction::TYPE);
    sa.setInterest(SoSearchAction::ALL);
    sa.apply(root);
    SoPathList plist = sa.getPaths();
    int len = plist.getLength();
    cerr << len << " cameras found, removing\n";
    for (int i=0; i<len; i++){
      SoPath *path = (SoPath*) plist[i];
      SoGroup *parent = (SoGroup *)plist[i]->getNodeFromTail(1);
      SoCamera *delCam = (SoCamera *)(plist[i]->getTail());
      parent->removeChild(delCam);
    }
  }

  else {
  
    SoSearchAction sa;
    sa.setType(SoCamera::getClassTypeId());
    sa.setSearchingAll(FALSE); // don't look under off switches
    sa.apply(root);
    SoCamera *sceneCam = NULL;
    SoFullPath *fullPath = NULL;  
    if (sa.getPath()){
      sceneCam = (SoCamera *)((SoFullPath *)sa.getPath())->getTail();
      fullPath = (SoFullPath *) sa.getPath();
    }
  
    // If camera exists, and is of a known type, move it to top of
    // scene root. If it's of unknown type, remove it and create
    // a perspective camera at scene root.
    //
    // We do this because the ExaminerViewer attaches to and modifies
    // the first camera (if it exists), modifying the scenegraph.
    //
    // Since we want to detect changes to the scenegraph, not including
    // view/camera changes, this is bad. Therefore we move the camera outside
    // of the scenegraph.
    if (sceneCam){

      // find parent node of camera
      SoGroup *parent = (SoGroup *)fullPath->getNodeFromTail(1);

      // identify camera type and if it's unknown,
      // create a new camara.
      if (sceneCam->isOfType(SoOrthographicCamera::getClassTypeId())){
	cerr << "Orthographic camera found...";
	newCam = sceneCam;
      }
      else if (sceneCam->isOfType(SoPerspectiveCamera::getClassTypeId())){
	cerr << "Perspective camera found...";
	newCam = sceneCam;
      }
      else {
	cerr << "Unknown camera type found, creating perspective camera...";

	// copy common stuff
	newCam = new SoPerspectiveCamera();
	newCam->viewportMapping = sceneCam->viewportMapping.getValue();
	newCam->position = sceneCam->position.getValue();
	newCam->orientation = sceneCam->orientation.getValue();
	newCam->aspectRatio = sceneCam->aspectRatio.getValue();
	newCam->focalDistance = sceneCam->focalDistance.getValue();
      }

      // remove the original camera
      newCam->ref();
      parent->removeChild(sceneCam);
      newCam->unrefNoDelete();
    }

    // otherwise create a camera
    else { 
      newCam = new SoPerspectiveCamera();
      newCam->viewAll(root, SbViewportRegion());
    }
  }

  // add camera to scene
  root->insertChild(newCam, 0);

  // create headlight
  //
  // headlightGroup - we have a rotation which keeps the headlight
  // moving whenever the camera moves,  and a reset xform so
  // that the rest of the scene is not affected by the first rot.
  // these leaves the direction field in the headlight open for the
  // user to edit, allowing for the direction to change w.r.t. the camera.
  //

  SoGroup *headlightGroup  = new SoGroup(3);
  SoTransform *headlightRot = new SoTransform();
  headlightRot->rotation.setValue(newCam->orientation.getValue());
  SoDirectionalLight *headlightNode = new SoDirectionalLight;
  headlightGroup->ref();
  headlightGroup->addChild(headlightRot);
  headlightGroup->addChild(headlightNode);
  headlightGroup->addChild(new SoResetTransform);
  headlightNode->direction.setValue(SbVec3f(.2, -.2, -.9797958971));
  root->insertChild(headlightGroup, 1);
}


void renderToFile(){
 
  cerr << "Rendering scene to file " << outFile << "..." << endl;
    
  // code taken from Open Inventor Mentor book p.221
  // create a viewport to render into
  SbViewportRegion myViewport;
  myViewport.setWindowSize(imgDim[0], imgDim[1]);

  // configure the renderer
  SoOffscreenRenderer *myRenderer = new SoOffscreenRenderer(myViewport);
  SoGLRenderAction *myAction = new SoGLRenderAction(myViewport);
  myAction->setSmoothing(useSmoothing);
  myAction->setNumPasses(numSamples);
  myRenderer->setGLRenderAction(myAction);

  // set options (JUNK CODE?)
//   SoGLRenderAction *myGLaction = myRenderer->getGLRenderAction();
//   if (!myGLaction){
//     cerr << "Error acquiring render action.\n";
//     exit(5);
//   }
//   myGLaction->setSmoothing(useSmoothing);
//   myGLaction->setNumPasses(numSamples);
  
  // render the scene
  SoNode *root = renderQuit ? scene : myViewer->getSceneManager()->getSceneGraph();
  if (root->isOfType(SoGroup::getClassTypeId()) && renderQuit)
      setCameraAndHeadlight((SoSeparator *)root, cameraFile);
  if (!myRenderer->render(root)){
    cerr << "Error rendering scene.\n";
    exit(2);
    delete myRenderer;
  }


  // back up old file if it exists
  struct stat file_stat;
  if (stat(outFile, &file_stat)==0){

    // find unused filename
    char *fn = new char[strlen(outFile)+100];
    int i;
    for (i=1;;i++){

      // insert number before extension
      char *lastPeriod = strrchr(outFile, '.');
      if (lastPeriod){
	int len = lastPeriod - outFile;
	strncpy(fn, outFile, len);
	sprintf(fn+len, ".%03d%s", i, lastPeriod);
      }

      // or afterwards if no extension found
      else 
	sprintf(fn, "%s.%03d", outFile, i);
      
      if (stat(fn, &file_stat))
	break;
    }

    // warn if many backups
    if (i>50)
      cerr << "Warning: excessive backups. You should probably delete some."
	   << endl;

    // copy file
    char *f = new char[strlen(outFile)*2+60];
    sprintf(f, "mv %s %s", outFile, fn);
    cerr << "Backing up old file..." << flush;
    cerr << f << endl;
    system(f);
      
    delete [] fn;
    delete [] f;
  }
  
    
  // open image file
  char tempName[50] = "";
  strncpy(tempName, "/tmp/renderToFile.eps.XXXXXX", sizeof tempName);
  FILE *fp;
  if ( mkstemp(tempName) == -1 || (fp = fopen(tempName, "w")) == NULL){
    cerr << "Could not open file " << outFile << " for writing: "
	 << strerror(errno) << "\n";
    exit(3);
  }

  // generate image
  //myRenderer->writeToRGB(fp);
  SbBool ok = myRenderer->writeToPostScript(fp);
  fclose(fp);

  if (!ok){
    cerr << "Error rendering.\n";
    exit(4);
  }

  // copy/convert file
  char *f = new char[strlen(outFile)+strlen(tempName)+60];
  sprintf(f, "convert %s %s", tempName, outFile);
  if (system(f) != 0)
    cerr << "Error converting file " << tempName << " to " << outFile << "!\n";
  else
    cerr << "Created " << outFile << "." << endl;

  // remove temporary file
  if (remove(tempName)){
    cerr << "Error removing temporary file " << tempName << ": "
	 << strerror(errno) << "\n";
  }

  delete myRenderer;

  return;
}

void keyEventHandler(void *gvp, SoEventCallback *callBack){
  const SoEvent *event = callBack->getEvent();

  // Render image to file when 'R' is pressed
  if (SO_KEY_PRESS_EVENT(event, R))
    renderToFile();
}

// static void mymainloop(){

//   XtAppContext context = SoXt::getAppContext();
//   XEvent event;
//   int i=0;

// //     for (;;) {
//   SoXt::nextEvent(context, &event);
//   SoXt::dispatchEvent(&event);


// }


int main(int argc, char *argv[]){

  // read options
  while (1){
    char c = getopt(argc, argv, "c:hld:s:o:r");

    if (c == -1)
      break;

    switch (c){

    case 'h':
       cerr << "Usage: " << argv[0]
	 << " [-h] [-dWxH] [-o outfile.rgb] [-l] [-sN] [-r] [scene.iv]\n"
	 << "\n"
	 << "Loads .iv scene file in an SoExaminerViewer and renders\n"
	 << "an image file in RGB format.\n"
	 << "\n"
	 << " -h              this help message\n"
         << " -c scene.iv    use first camera from inventor scene scene.iv\n"
	 << " -dWxH           resolution of image\n"
	 << " -o outfile.rgb  name of output rgb file (image.rgb)\n"
	 << " -l              use line smoothing\n"
	 << " -sN             use N samples for super sampling (1)\n"
	 << " -r              render straight to file without viewer.\n";
	 
      exit(1);
      break;

    case 'd':
      imgDim[0] = atoi(optarg);
      imgDim[1] = atoi(strchr(optarg, 'x')+1);
      if (imgDim[0] <= 0 || imgDim[1] <= 0){
	cerr << "Invalid image dimensions: " << optarg << endl;
	exit(1);
      }
      break;

    case 'c':
      cameraFile = optarg;
      break;
      
    case 'o':
      outFile = optarg;
      break;
      
    case 'l':
      useSmoothing = TRUE;
      break;
      
    case 's':
      numSamples = atoi(optarg);
      break;

     case 'r':
       renderQuit = TRUE;
       break;
       
     default:
       break;
    }
  }

  // read input filename
  if (optind<argc){
    inFile = argv[optind++];
  }
  
  while (optind<argc){
    cerr << "Unknown option: " << argv[optind++] << "\n";
  }
  
  // initialize inventor
  if (renderQuit)
    SoDB::init();
  else
    SoXt::init(argv[0]);

  // open scene
  SoInput *sceneIn = new SoInput();
  if (!sceneIn->openFile(inFile)){
    cerr << "Unable to open scene file: " << inFile << endl;
    exit(1);
  }

  // read scene
  scene = SoDB::readAll(sceneIn);
  if (!scene){
    cerr << "Error: scene separator is null!" << endl;
    exit(1);
  }
  scene->ref();   // ref count is initially zero (why I dunno...)
  delete sceneIn;

  if (!renderQuit){
    // Set up keyboard
    SoEventCallback *callBack = new SoEventCallback();
    callBack->addEventCallback(SoKeyboardEvent::getClassTypeId(),
			       keyEventHandler);
    scene->addChild(callBack);
    cerr << "\nPress ESC, then R to render to file.\n" << endl;

    // This returns a main window to use.
    // If unsuccessful, exit.
    Widget myWindow = SoXt::init(argv[0]); // pass the app name
    if (myWindow == NULL) exit(1);

    // Set up viewer
    myViewer = new SoXtExaminerViewer(myWindow);
    myViewer->setSceneGraph(scene);
    myViewer->setTitle("Scene Viewer");
    myViewer->setAntialiasing(useSmoothing, numSamples); // optional multisampling
    myViewer->setSize(SbVec2s(imgDim[0], imgDim[1]));
    myViewer->show();
    const SbVec2s &size = myViewer->getViewportRegion().getWindowSize();
    cerr << "Window size: " << size[0] << "x" << size[1] << endl;
    
    SoXt::show(myWindow); // Display main window

    // find window size
    const SbVec2s &imagePixSize = myViewer->getViewportRegion().getViewportSizePixels();
    cerr << "Window size: "
	 << imagePixSize[0] << "x" << imagePixSize[1] << endl;
    
    SoXt::mainLoop();	// Main Inventor event loop
  }
  else
    renderToFile();
  
  return(0);
}
